# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2008 rule R1117

    submodule-stmt is SUBMODULE ( parent-identifier ) submodule-name

'''
import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2008 import Submodule_Stmt


def test_simple(f2008_create):
    ''' Test the parsing of a submodule statement'''
    result = Submodule_Stmt("submodule (id) name")
    assert str(result) == "SUBMODULE (id) name"


def test_simple_error1(f2008_create):
    ''' Test the parsing of a submodule statement'''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submod (id) name")
    assert "Submodule_Stmt: 'submod (id) name" in str(excinfo.value)


def test_simple_error2(f2008_create):
    ''' Test the parsing of a submodule statement'''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule name")
    assert "Submodule_Stmt: 'submodule name'" in str(excinfo.value)


def test_simple_error3(f2008_create):
    ''' Test the parsing of a submodule statement'''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule () name")
    assert "Submodule_Stmt: 'submodule () name'" in str(excinfo.value)


def test_simple_error4(f2008_create):
    ''' Test the parsing of a submodule statement'''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule (id)")
    assert "Submodule_Stmt: 'submodule (id)'" in str(excinfo.value)


def test_simple_error5(f2008_create):
    ''' Test the parsing of a submodule statement'''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule name (id)")
    assert "Submodule_Stmt: 'submodule name (id)'" in str(excinfo.value)


def test_simple_error6(f2008_create):
    ''' Test the parsing of a submodule statement'''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule (id) (name)")
    assert "Submodule_Stmt: 'submodule (id) (name)'" in str(excinfo.value)


def test_simple_error7(f2008_create, monkeypatch):
    '''Test the parsing of a submodule statement when there is a single
    right hand bracket. The error generated here is unreachable if
    splitparen works correctly so we need to monkeypatch.

    '''

    monkeypatch.setattr("fparser.common.splitline.splitparen",
                        lambda x: ["", "id)", "name"])
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule id) name")
    assert "Submodule_Stmt: 'submodule id) name'" in str(excinfo.value)


def test_simple_error8(f2008_create, monkeypatch):
    '''Test the parsing of a submodule statement when there is a single
    left hand bracket. The error generated here is unreachable if
    splitparen works correctly so we need to monkeypatch.

    '''

    monkeypatch.setattr("fparser.common.splitline.splitparen",
                        lambda x: ["", "(id", "name"])
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule (id name")
    assert "Submodule_Stmt: 'submodule (id name'" in str(excinfo.value)


def test_splitparen_error(monkeypatch):
    '''Test that if the first argument of the splitparen is not empty then
    an error is returned.

    '''

    monkeypatch.setattr("fparser.common.splitline.splitparen",
                        lambda x: ["XXX", "", ""])
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule (id) name")
    assert "Submodule_Stmt: 'submodule (id) name'" in str(excinfo.value)


def test_simple_error9(f2008_create):
    ''' Test the parsing of a submodule statement'''
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Submodule_Stmt("submodule (id) name :")
    assert "Submodule_Stmt: 'submodule (id) name :'" in str(excinfo.value)



#!/usr/bin/env python
# Modified work Copyright (c) 2017 Science and Technology Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from __future__ import print_function
import logging
from fparser.scripts.script_options import set_parse_options
import fparser.common.sourceinfo

logging.basicConfig()

# START UPDATE SYS.PATH ###
# END UPDATE SYS.PATH ###
try:
    from iocbio.optparse_gui import OptionParser
except ImportError:
    from optparse import OptionParser


def runner(parser, options, args):
    from fparser.common.readfortran import FortranFileReader
    from fparser.one.parsefortran import FortranParser
    for filename in args:
        reader = FortranFileReader(filename)
        if options.mode != 'auto':
            mode = fparser.common.sourceinfo\
                   .FortranFormat.from_mode(options.mode)
            reader.format.set_mode(mode)
        parser = FortranParser(reader)
        parser.parse()
        parser.analyze()
        if options.task == 'show':
            print(parser.block.torepr(4))
        elif options.task == 'none':
            pass
        else:
            raise NotImplementedError(repr(options.task))


def main():
    parser = OptionParser()
    set_parse_options(parser)
    if hasattr(parser, 'runner'):
        parser.runner = runner
    options, args = parser.parse_args()
    runner(parser, options, args)
    return


if __name__ == "__main__":
    main()



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R304 : This file tests support for the Name
class.

'''

import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2003 import Name

# Test the name class R304


def test_valid_names():
    '''Test the appropriate set of valid names. Note, '$' is not valid in
    the standard but is supported by some compilers so we allow it
    here.

    '''
    for name in ["a", "abcde", "a_", "a1", "  a  ", "a$"]:
        result = Name(name)
        assert str(result) == name.strip()


def test_invalid_names():
    '''Test invalid names. There are too many to try all invalid
    characters so just do a subset.

    '''
    for name in ["1", "_", "1a", "_a", "a a", "", " ", "a!", "a@", "a#",
                 "a%", "a^", "a&", "a*", "a(", "a)"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Name(name)
        assert "Name: '{0}'".format(name) in str(excinfo.value)



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R404 : This file tests the support for the
kind selector.

'''

import pytest
from fparser.two.utils import InternalError
from fparser.api import get_reader
from fparser.two.Fortran2003 import Kind_Selector, Program

# match() 'kind=' is present


def test_kind(f2003_create):
    '''Test that a kind selector with 'kind=' present can be parsed
    successfully.

    '''
    reader = get_reader("(KIND=some_kind)")
    ast = Kind_Selector(reader)
    assert "(KIND = some_kind)" in str(ast)


def test_kind_case(f2003_create):
    '''Test that a kind selector with 'kind=' present with mixed case can
    be parsed successfully.

    '''
    reader = get_reader("(KiNd=some_kind)")
    ast = Kind_Selector(reader)
    assert "(KIND = some_kind)" in str(ast)


def test_kind_spaces(f2003_create):
    '''Test that a kind selector with 'kind=' present and spaces can be
    parsed successfully.

    '''
    reader = get_reader("  (  kind  =  some_kind  )  ")
    ast = Kind_Selector(reader)
    assert "(KIND = some_kind)" in str(ast)

# match() 'kind=' is not present


def test_nokind(f2003_create):
    '''Test that a kind selector with 'kind=' not present can be parsed
    successfully. Note, the parser automatically adds-in kind for the
    output.

    '''
    reader = get_reader("(some_kind)")
    ast = Kind_Selector(reader)
    assert "(KIND = some_kind)" in str(ast)


def test_nokind_spaces(f2003_create):
    '''Test that a kind selector with 'kind=' not present and spaces can
    be parsed successfully. Note, the parser automatically adds-in
    kind for the output.

    '''
    reader = get_reader("  (  some_kind  )  ")
    ast = Kind_Selector(reader)
    assert "(KIND = some_kind)" in str(ast)


def test_nokind_short(f2003_create):
    '''Test that a kind selector with 'kind=' not present and a name less
    than 6 characters can be parsed successfully. This is important as
    there is a pattern match for 'kind=' so we need to test we don't
    accidentally access outside of the string. Note, the parser
    automatically adds-in kind for the output.

    '''
    reader = get_reader("(a)")
    ast = Kind_Selector(reader)
    assert "(KIND = a)" in str(ast)


def test_nokind_kind_function(f2003_create):
    '''Test that a kind selector with 'kind=' not present and a kind
    function can be parsed successfully. This is a test as there is a
    risk that the pattern match will think the kind function is the
    start of 'kind='. Note, the parser automatically adds-in kind for
    the output.

    '''
    reader = get_reader("(kind(kind_id))")
    ast = Kind_Selector(reader)
    assert "(KIND = KIND(kind_id))" in str(ast)

# match() *n extension


def test_size_extension(f2003_create):
    '''Test that a kind selector with the '*number' format e.g. integer*4
    can be parsed successfully. This is an extension to the standard
    but is well used and is supported in mainstream compilers so is
    also supported here. Note, we can't use a reader as it will not
    read '*2' so we pass the string directly.

    '''
    ast = Kind_Selector("*2")
    assert "*2" in str(ast)


def test_size_extension_spaces(f2003_create):
    '''Test that a kind selector with the '*number' format e.g. integer*4
    can be parsed successfully if there are spaces between the * and
    the number. This is an extension to the standard but is well used
    and is supported in mainstream compilers so is also supported
    here. Note, we can't use a reader as it will not read '*2' so we
    pass the string directly.

    '''
    ast = Kind_Selector("  *  2  ")
    assert "*2" in str(ast)

# match() Syntax errors


def test_error_left_bracket(f2003_create):
    '''Test that None is returned if the left bracket is missing.'''
    reader = get_reader("KIND = some_kind)")
    ast = Kind_Selector(reader)
    assert not ast


def test_error_right_bracket(f2003_create):
    '''Test that None is returned if the right bracket is missing.'''
    reader = get_reader("(KIND = some_kind")
    ast = Kind_Selector(reader)
    assert not ast


def test_error_kind_name(f2003_create):
    '''Test that None is returned if the name 'kind' is incorrect.'''
    reader = get_reader("(KINE = some_kind")
    ast = Kind_Selector(reader)
    assert not ast


def test_error_kind_equals(f2003_create):
    '''Test that None is returned if the '=' after 'kind' is not there.'''
    reader = get_reader("(KIND some_kind")
    ast = Kind_Selector(reader)
    assert not ast

# match() Internal errors


def test_error_star_only(f2003_create):
    '''Test that an InternalError is raised if there are less than two
    characters in the string as valid input must have at least two.

    '''
    with pytest.raises(InternalError) as excinfo:
        _ = Kind_Selector("*")
    assert "too short to be valid" in str(excinfo.value)


def test_error_empty(f2003_create):
    '''Test that an InternalError is raised if the input is empty.

    '''
    with pytest.raises(InternalError) as excinfo:
        _ = Kind_Selector("")
    assert "too short to be valid" in str(excinfo.value)


def test_error_none(f2003_create):
    '''Test that an InternalError is raised if the input is None.'''
    with pytest.raises(InternalError) as excinfo:
        _ = Kind_Selector(None)
    assert "method match() is None" in str(excinfo.value)

# tostr() Errors


def test_tostr_error_nitems(f2003_create, monkeypatch):
    '''Test that the appropriate exception is raised if the number of
    elements in the self.items list is >3 or <2.

    '''
    ast = Kind_Selector("*8")
    monkeypatch.setattr(ast, "items", ["*"])
    with pytest.raises(InternalError) as excinfo:
        _ = str(ast)
    assert ("tostr() has '1' items, but expecting 2 or 3"
            in str(excinfo.value))

# misc


@pytest.mark.xfail(reason="The x gets lost when there is a space in "
                   "front of it.")
def test_to_be_moved(f2003_create):
    '''Show that an example with a space before a kind specification
    fails. This is not the fault of this class but I'm including it
    here for the moment until the reason for the error is fixed.

    '''
    reader = get_reader('''\
      subroutine test()
        integer( x) y
      end subroutine
      ''')
    ast = Program(reader)
    assert ("SUBROUTINE test\n"
            "  integer( x) y\n"
            "END SUBROUTINE") in str(ast)



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Test Fortran 2003 rule R451 : This file tests the support for specific
binding and passed-object dummy argument within the type-bound procedure
part of a derived type.
'''

import pytest
from fparser.two.utils import NoMatchError, InternalError
from fparser.two.Fortran2003 import Specific_Binding


def test_valid(f2003_create):
    '''Test that valid Specific_Binding statements are parsed
    correctly.

    '''

    # simple statement
    obj = Specific_Binding('procedure sub')
    assert str(obj) == "PROCEDURE sub"

    # simple statement spaces
    obj = Specific_Binding('  procedure  sub  ')
    assert str(obj) == "PROCEDURE sub"

    # multi-case procedure keyword
    obj = Specific_Binding('PrOcEdUrE sub')
    assert str(obj) == "PROCEDURE sub"

    # simple statement with ::
    obj = Specific_Binding('procedure :: sub')
    assert str(obj) == "PROCEDURE :: sub"

    # simple statement with :: no space
    obj = Specific_Binding('procedure::sub')
    assert str(obj) == "PROCEDURE :: sub"

    # statement with procedure name
    obj = Specific_Binding('procedure :: sub => boat')
    assert str(obj) == "PROCEDURE :: sub => boat"

    # statement with procedure name no space
    obj = Specific_Binding('procedure::sub=>boat')
    assert str(obj) == "PROCEDURE :: sub => boat"

    # statement with interface name
    obj = Specific_Binding('procedure ( interface ) sub')
    assert str(obj) == "PROCEDURE(interface) sub"

    # statement with interface name no space
    obj = Specific_Binding('procedure(interface)::sub')
    assert str(obj) == "PROCEDURE(interface) :: sub"

    # statement with binding list
    obj = Specific_Binding('procedure , PASS :: sub')
    assert str(obj) == "PROCEDURE, PASS :: sub"

    # statement with binding list no space
    obj = Specific_Binding('procedure,PASS::sub')
    assert str(obj) == "PROCEDURE, PASS :: sub"

    # statement with multi-binding list
    obj = Specific_Binding('procedure , PASS , PUBLIC, deferred :: sub')
    assert str(obj) == "PROCEDURE, PASS, PUBLIC, DEFERRED :: sub"

    # statement with multi-binding list no space
    obj = Specific_Binding('procedure,PASS,PUBLIC,DEFERRED::sub')
    assert str(obj) == "PROCEDURE, PASS, PUBLIC, DEFERRED :: sub"

    # statement with interface name and binding list
    obj = Specific_Binding('procedure ( interface ) , PASS :: sub')
    assert str(obj) == "PROCEDURE(interface), PASS :: sub"

    # statement with interface name and binding list no space
    obj = Specific_Binding('procedure(interface),PASS::sub')
    assert str(obj) == "PROCEDURE(interface), PASS :: sub"

    # statement with binding list and procedure name
    obj = Specific_Binding('procedure , PASS :: sub => boat')
    assert str(obj) == "PROCEDURE, PASS :: sub => boat"

    # statement with binding list and procedure name no space
    obj = Specific_Binding('procedure,PASS::sub=>boat')
    assert str(obj) == "PROCEDURE, PASS :: sub => boat"


def test_invalid():
    ''' Test that parsing invalid Fortran syntax for
    Specific_Binding statements raises an appropriate error. '''

    # includes tests for breaking C456 and C457
    for string in ["",
                   "  ",
                   "procedure  ",
                   "procedure ::  ",
                   "procedur sub",
                   "proceduresub",
                   "procedure : sub",
                   "procedure sub => boat",  # C456
                   "procedure :: sub =>",
                   "procedure (interface sub",
                   "procedure interface) sub",
                   "procedure (interface) :: sub => boat",  # C457
                   "procedure pass sub",
                   "procedure, pass sub",
                   "procedure, pass public sub",
                   "procedure pass, public sub",
                   "procedure pass :: sub",
                   "procedure sub sub2",
                   "procedure :: sub = boat",
                   "procedure :: sub > boat",
                   "procedure :: sub => boat boat2"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Specific_Binding(string)
        assert "Specific_Binding: '{0}'".format(string) in str(excinfo.value)


def test_tostr_invalid(f2003_create, monkeypatch):
    ''' Test that invalid input raises an exception '''

    # test internal error in tostr() when the items list is not the
    # expected size
    obj = Specific_Binding("procedure hello")
    monkeypatch.setattr(obj, "items", ["procedure"])
    with pytest.raises(InternalError) as excinfo:
        _ = str(obj)
    assert "tostr() has '1' items, but expecting 5" in str(excinfo.value)



# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

# Original author: Pearu Peterson <pearu@cens.ioc.ee>
# First version created: May 2006

"""
Test parsing single Fortran lines.

"""

import pytest
from fparser.one.block_statements import Allocatable, Allocate, ArithmeticIf, \
     AssignedGoto, Assign, Assignment, Asynchronous, Backspace, Bind, Call, \
     CallProtoArgument, CallStatement, Check, Close, Common, \
     ComputedGoto, Contains, Continue, Cycle, Data, Deallocate, Depend, \
     Dimension, Else, ElseIf, ElseWhere, Endfile, Entry, Enumerator, \
     Equivalence, Exit, External, FinalBinding, Flush, Format, ForallStmt, \
     FortranName, GenericBinding, Goto, Implicit, Import, Inquire, \
     Intent, Intrinsic, ModuleProcedure, Nullify, Open, Optional, Parameter, \
     Pointer, PointerAssignment, Pause, Print, Private, Protected, Public, \
     Read, Return, Rewind, Save, Sequence, SpecificBinding, Stop, Target, \
     Threadsafe, Use, Value, Volatile, Wait, WhereStmt, Write
import fparser.common.sourceinfo
from fparser.one.typedecl_statements import Byte, Character, Complex, \
     DoubleComplex, DoublePrecision, Integer, Logical, Real

from fparser.common.readfortran import FortranStringReader
from fparser.common.utils import AnalyzeError, ParseError


def parse(cls, line, label='', isfree=True, isstrict=False):
    '''Tries to parse a Fortran line using the given class cls.
    If successful, it then converts the parsed statement back to
    a string. If isstrict is false, it will then try to parse
    this string again (recursively calling itself, with isstrict
    set to true) and make sure that the re-parsed string is
    identical to the input.
    It returns the string representation of the parsed input line.
    '''

    if label:
        line = label + ' : ' + line
    reader = FortranStringReader(line)
    reader.set_format(fparser.common.sourceinfo.FortranFormat(isfree,
                                                              isstrict))
    item = next(reader)
    if not cls.match(item.get_line()):
        raise ValueError('%r does not match %s pattern' % (line, cls.__name__))
    stmt = cls(item, item)
    if stmt.isvalid:
        # Check that we can successfully parse the string representation
        # of the parsed object
        stmt_string = str(stmt)
        if not isstrict:
            reparsed_stmt_string = parse(cls, stmt_string, isstrict=True)
            if stmt_string != reparsed_stmt_string:
                raise ValueError('Failed to parse %r with %s pattern in Pyf '
                                 'mode, got %r' %
                                 (stmt_string, cls.__name__,
                                  reparsed_stmt_string))
        return stmt_string
    raise ValueError('parsing %r with %s pattern failed' %
                     (line, cls.__name__))


def test_assignment():
    assert parse(Assignment, 'a=b') == 'a = b'
    assert parse(PointerAssignment, 'a=>b') == 'a => b'
    assert parse(Assignment, 'a (2)=b(n,m)') == 'a(2) = b(n,m)'
    assert parse(Assignment, 'a % 2(2,4)=b(a(i))') == 'a%2(2,4) = b(a(i))'


def test_assign():
    assert parse(Assign, 'assign 10 to a') == 'ASSIGN 10 TO a'


def test_call():
    assert parse(Call, 'call a') == 'CALL a'
    assert parse(Call, 'call a()') == 'CALL a'
    assert parse(Call, 'call a(1)') == 'CALL a(1)'
    assert parse(Call, 'call a(1,2)') == 'CALL a(1, 2)'
    assert parse(Call, 'call a % c ( n , a+1 )') == 'CALL a % c(n, a+1)'


def test_goto():
    assert parse(Goto, 'go to 19') == 'GO TO 19'
    assert parse(Goto, 'goto 19') == 'GO TO 19'
    assert parse(ComputedGoto, 'goto (1, 2 ,3) a+b(2)') == \
        'GO TO (1, 2, 3) a+b(2)'
    assert parse(ComputedGoto, 'goto (1, 2 ,3) , a+b(2)') == \
        'GO TO (1, 2, 3) a+b(2)'
    assert parse(AssignedGoto, 'goto a') == 'GO TO a'
    assert parse(AssignedGoto, 'goto a ( 1 )') == 'GO TO a (1)'
    assert parse(AssignedGoto, 'goto a ( 1 ,2)') == 'GO TO a (1, 2)'


def test_continue():
    assert parse(Continue, 'continue') == 'CONTINUE'


def test_return():
    assert parse(Return, 'return') == 'RETURN'
    assert parse(Return, 'return a') == 'RETURN a'
    assert parse(Return, 'return a+1') == 'RETURN a+1'
    assert parse(Return, 'return a(c, a)') == 'RETURN a(c, a)'


def test_stop():
    assert parse(Stop, 'stop') == 'STOP'
    assert parse(Stop, 'stop 1') == 'STOP 1'
    assert parse(Stop, 'stop "a"') == 'STOP "a"'
    assert parse(Stop, 'stop "a b"') == 'STOP "a b"'


def test_print():
    assert parse(Print, 'print*') == 'PRINT *'
    assert parse(Print, 'print "a b( c )"') == 'PRINT "a b( c )"'
    assert parse(Print, 'print 12, a') == 'PRINT 12, a'
    assert parse(Print, 'print 12, a , b') == 'PRINT 12, a, b'
    assert parse(Print, 'print 12, a(c,1) , b') == 'PRINT 12, a(c,1), b'


def test_read():
    assert parse(Read, 'read ( 10 )') == 'READ (10)'
    assert parse(Read, 'read ( 10 ) a ') == 'READ (10) a'
    assert parse(Read, 'read ( 10 ) a , b') == 'READ (10) a, b'
    assert parse(Read, 'read *') == 'READ *'
    assert parse(Read, 'read 12') == 'READ 12'
    assert parse(Read, 'read "a b"') == 'READ "a b"'
    assert parse(Read, 'read "a b",a') == 'READ "a b", a'
    assert parse(Read, 'read * , a') == 'READ *, a'
    assert parse(Read, 'read "hey a" , a') == 'READ "hey a", a'
    assert parse(Read, 'read * , a  , b') == 'READ *, a, b'
    assert parse(Read, 'read ( unit  =10 )') == 'READ (UNIT = 10)'


def test_write():
    assert parse(Write, 'write ( 10 )') == 'WRITE (10)'
    assert parse(Write, 'write ( 10 , a )') == 'WRITE (10, a)'
    assert parse(Write, 'write ( 10 ) b') == 'WRITE (10) b'
    assert parse(Write, 'write ( 10 ) a(1) , b+2') == \
        'WRITE (10) a(1), b+2'
    assert parse(Write, 'write ( unit=10 )') == 'WRITE (UNIT = 10)'


def test_flush():
    assert parse(Flush, 'flush 10') == 'FLUSH (10)'
    assert parse(Flush, 'flush (10)') == 'FLUSH (10)'
    assert parse(Flush, 'flush (UNIT = 10)') == 'FLUSH (UNIT = 10)'
    assert parse(Flush, 'flush (10, err=  23)') == 'FLUSH (10, ERR = 23)'


def test_wait():
    assert parse(Wait, 'wait(10)') == 'WAIT (10)'
    assert parse(Wait, 'wait(10,err=129)') == 'WAIT (10, ERR = 129)'


def test_contains():
    assert parse(Contains, 'contains') == 'CONTAINS'


def test_allocate():
    ''' Tests for various forms of ALLOCATE statement '''
    assert parse(Allocate, 'allocate (a)') == 'ALLOCATE (a)'
    assert parse(Allocate, 'allocate (a, stat=b)') == \
        'ALLOCATE (a, STAT = b)'
    assert parse(Allocate, 'allocate (a,b(:1))') == 'ALLOCATE (a, b(:1))'
    assert parse(Allocate, 'allocate (real(8)::a)') == \
        'ALLOCATE (REAL(KIND=8) :: a)'
    # With a type specifier
    assert parse(Allocate, 'allocate (real :: a(8))') == \
        'ALLOCATE (REAL :: a(8))'
    assert parse(Allocate, 'allocate (real(kind=wp) :: a(8))') == \
        'ALLOCATE (REAL(KIND=wp) :: a(8))'
    assert parse(Allocate, 'allocate (a_type :: a)') == \
        'ALLOCATE (a_type :: a)'
    with pytest.raises(ParseError) as err:
        parse(Allocate, 'allocate(not valid :: a)')
    assert "Unrecognised type-specification in ALLOCATE statement" in \
        str(err.value)


def test_deallocate():
    assert parse(Deallocate, 'deallocate (a)') == 'DEALLOCATE (a)'
    assert parse(Deallocate, 'deallocate (a, stat=b)') == \
        'DEALLOCATE (a, STAT = b)'


def test_moduleprocedure():
    ''' Test  [ MODULE ] PROCEDURE [::] <procedure-name-list> '''

    assert parse(ModuleProcedure, 'Procedure a') == 'MODULE PROCEDURE a'
    assert parse(ModuleProcedure, 'procedure a , b') == \
        'MODULE PROCEDURE a, b'
    assert parse(ModuleProcedure, 'procedure :: a ') == \
        'MODULE PROCEDURE a'
    assert parse(ModuleProcedure, 'procedure :: a , b') == \
        'MODULE PROCEDURE a, b'
    assert parse(ModuleProcedure, 'ModuleProcedure a') == \
        'MODULE PROCEDURE a'
    assert parse(ModuleProcedure, 'module procedure a , b') == \
        'MODULE PROCEDURE a, b'
    assert parse(ModuleProcedure, 'module procedure :: a ') == \
        'MODULE PROCEDURE a'
    assert parse(ModuleProcedure, 'module procedure :: a , b') == \
        'MODULE PROCEDURE a, b'
    assert parse(ModuleProcedure, 'moduleprocedure::a,b') == \
        'MODULE PROCEDURE a, b'


def test_access():
    assert parse(Public, 'Public') == 'PUBLIC'
    assert parse(Public, 'public a') == 'PUBLIC a'
    assert parse(Public, 'public :: a') == 'PUBLIC a'
    assert parse(Public, 'public a,b,c') == 'PUBLIC a, b, c'
    assert parse(Public, 'public :: a(:,:)') == 'PUBLIC a(:,:)'
    assert parse(Private, 'private') == 'PRIVATE'
    assert parse(Private, 'private :: a') == 'PRIVATE a'


def test_close():
    assert parse(Close, 'close (12)') == 'CLOSE (12)'
    assert parse(Close, 'close (12, err=99)') == 'CLOSE (12, ERR = 99)'
    assert parse(Close, 'close (12, status = a(1,2))') == \
        'CLOSE (12, STATUS = a(1,2))'


def test_class():
    ''' Check that we correctly parse and generate a class declaration '''
    from fparser.one.typedecl_statements import Class
    assert parse(Class, 'class(runtime_constants_type) :: a') == \
        "CLASS(runtime_constants_type) a"


def test_cycle():
    assert parse(Cycle, 'cycle') == 'CYCLE'
    assert parse(Cycle, 'cycle ab') == 'CYCLE ab'


def test_rewind():
    assert parse(Rewind, 'rewind 1') == 'REWIND (1)'
    assert parse(Rewind, 'rewind (1)') == 'REWIND (1)'
    assert parse(Rewind, 'rewind (1, err =  123)') == \
        'REWIND (1, ERR = 123)'


def test_backspace():
    assert parse(Backspace, 'backspace 1') == 'BACKSPACE (1)'
    assert parse(Backspace, 'backspace (1)') == 'BACKSPACE (1)'
    assert parse(Backspace, 'backspace (1, err =  123)') == \
        'BACKSPACE (1, ERR = 123)'


def test_endfile():
    assert parse(Endfile, 'endfile 1') == 'ENDFILE (1)'
    assert parse(Endfile, 'endfile (1)') == 'ENDFILE (1)'
    assert parse(Endfile, 'endfile (1, err =  123)') == \
        'ENDFILE (1, ERR = 123)'


def test_open():
    assert parse(Open, 'open (1)') == 'OPEN (1)'
    assert parse(Open, 'open (1, err =  123)') == 'OPEN (1, ERR = 123)'


def test_format():
    assert parse(Format, '1 format ()') == '1 FORMAT ()'
    assert parse(Format, '199 format (1)') == '199 FORMAT (1)'
    assert parse(Format, '2 format (1 , SS)') == '2 FORMAT (1, ss)'


def test_save():
    assert parse(Save, 'save') == 'SAVE'
    assert parse(Save, 'save :: a') == 'SAVE a'
    assert parse(Save, 'save a,b') == 'SAVE a, b'


def test_data():
    assert parse(Data, 'data a /b/') == 'DATA a / b /'
    assert parse(Data, 'data a , c /b/') == 'DATA a, c / b /'
    assert parse(Data, 'data a /b ,c/') == 'DATA a / b, c /'
    assert parse(Data, 'data a /b/ c,e /d/') == 'DATA a / b / c, e / d /'
    assert parse(Data, 'data a(1,2) /b/') == 'DATA a(1,2) / b /'
    assert parse(Data, 'data a /b, c(1)/') == 'DATA a / b, c(1) /'


def test_nullify():
    assert parse(Nullify, 'nullify(a)') == 'NULLIFY (a)'
    assert parse(Nullify, 'nullify(a  ,b)') == 'NULLIFY (a, b)'


def test_use():
    ''' Check that we can parse USE statement. '''
    assert parse(Use, 'use a') == 'USE a'
    assert parse(Use, 'use :: a') == 'USE a'
    assert parse(Use, 'use, intrinsic:: a') == 'USE, INTRINSIC :: a'
    assert parse(Use, 'use, non_intrinsic:: a') == 'USE, NON_INTRINSIC :: a'
    assert parse(Use, 'use, non_intrinsic:: a , only: b') == \
        'USE, NON_INTRINSIC :: a, ONLY: b'
    assert parse(Use, 'use :: a ,only: b') == 'USE a, ONLY: b'
    assert parse(Use, 'use :: a , only: b=>c') == 'USE a, ONLY: b=>c'
    assert parse(Use, 'use :: a , b=>c') == 'USE a, b=>c'
    assert parse(Use, 'use :: a , only: operator(+) , b') == \
        'USE a, ONLY: operator(+), b'


def test_exit():
    assert parse(Exit, 'exit') == 'EXIT'
    assert parse(Exit, 'exit ab') == 'EXIT ab'


def test_parameter():
    assert parse(Parameter, 'parameter (a = b(1,2))') == \
        'PARAMETER (a = b(1,2))'
    assert parse(Parameter, 'parameter (a = b(1,2) , b=1)') == \
        'PARAMETER (a = b(1,2), b=1)'


def test_equivalence():
    assert parse(Equivalence, 'equivalence (a , b)') == \
        'EQUIVALENCE (a, b)'
    assert parse(Equivalence, 'equivalence (a , b) , ( c, d(1) , g  )') == \
        'EQUIVALENCE (a, b), (c, d(1), g)'


def test_dimension():
    assert parse(Dimension, 'dimension a(b)') == 'DIMENSION a(b)'
    assert parse(Dimension, 'dimension::a(b)') == 'DIMENSION a(b)'
    assert parse(Dimension, 'dimension a(b)  , c(d)') == \
        'DIMENSION a(b), c(d)'
    assert parse(Dimension, 'dimension a(b,c)') == 'DIMENSION a(b,c)'


def test_target():
    assert parse(Target, 'target a(b)') == 'TARGET a(b)'
    assert parse(Target, 'target::a(b)') == 'TARGET a(b)'
    assert parse(Target, 'target a(b)  , c(d)') == 'TARGET a(b), c(d)'
    assert parse(Target, 'target a(b,c)') == 'TARGET a(b,c)'


def test_pointer():
    assert parse(Pointer, 'pointer a=b') == 'POINTER a=b'
    assert parse(Pointer, 'pointer :: a=b') == 'POINTER a=b'
    assert parse(Pointer, 'pointer a=b, c=d(1,2)') == \
        'POINTER a=b, c=d(1,2)'


def test_protected():
    assert parse(Protected, 'protected a') == 'PROTECTED a'
    assert parse(Protected, 'protected::a') == 'PROTECTED a'
    assert parse(Protected, 'protected a , b') == 'PROTECTED a, b'


def test_volatile():
    assert parse(Volatile, 'volatile a') == 'VOLATILE a'
    assert parse(Volatile, 'volatile::a') == 'VOLATILE a'
    assert parse(Volatile, 'volatile a , b') == 'VOLATILE a, b'


def test_value():
    assert parse(Value, 'value a') == 'VALUE a'
    assert parse(Value, 'value::a') == 'VALUE a'
    assert parse(Value, 'value a , b') == 'VALUE a, b'


def test_arithmeticif():
    assert parse(ArithmeticIf, 'if (a) 1,2,3') == 'IF (a) 1, 2, 3'
    assert parse(ArithmeticIf, 'if (a(1)) 1,2,3') == 'IF (a(1)) 1, 2, 3'
    assert parse(ArithmeticIf, 'if (a(1,2)) 1,2,3') == \
        'IF (a(1,2)) 1, 2, 3'


def test_intrinsic():
    assert parse(Intrinsic, 'intrinsic a') == 'INTRINSIC a'
    assert parse(Intrinsic, 'intrinsic::a') == 'INTRINSIC a'
    assert parse(Intrinsic, 'intrinsic a , b') == 'INTRINSIC a, b'


def test_inquire():
    assert parse(Inquire, 'inquire (1)') == 'INQUIRE (1)'
    assert parse(Inquire, 'inquire (1, err=123)') == \
        'INQUIRE (1, ERR = 123)'
    assert parse(Inquire, 'inquire (iolength=a) b') == \
        'INQUIRE (IOLENGTH = a) b'
    assert parse(Inquire, 'inquire (iolength=a) b  ,c(1,2)') == \
        'INQUIRE (IOLENGTH = a) b, c(1,2)'


def test_sequence():
    assert parse(Sequence, 'sequence') == 'SEQUENCE'


def test_external():
    assert parse(External, 'external a') == 'EXTERNAL a'
    assert parse(External, 'external::a') == 'EXTERNAL a'
    assert parse(External, 'external a , b') == 'EXTERNAL a, b'


def test_common():
    assert parse(Common, 'common a') == 'COMMON a'
    assert parse(Common, 'common a , b') == 'COMMON a, b'
    assert parse(Common, 'common a , b(1,2)') == 'COMMON a, b(1,2)'
    assert parse(Common, 'common // a') == 'COMMON a'
    assert parse(Common, 'common / name/ a') == 'COMMON / name / a'
    assert parse(Common, 'common / name/ a  , c') == \
        'COMMON / name / a, c'
    assert parse(Common, 'common / name/ a /foo/ c(1) ,d') == \
        'COMMON / name / a / foo / c(1), d'
    assert parse(Common, 'common / name/ a, /foo/ c(1) ,d') == \
        'COMMON / name / a / foo / c(1), d'


def test_optional():
    assert parse(Optional, 'optional a') == 'OPTIONAL a'
    assert parse(Optional, 'optional::a') == 'OPTIONAL a'
    assert parse(Optional, 'optional a , b') == 'OPTIONAL a, b'


def test_intent():
    assert parse(Intent, 'intent (in) a') == 'INTENT (IN) a'
    assert parse(Intent, 'intent(in)::a') == 'INTENT (IN) a'
    assert parse(Intent, 'intent(in) a , b') == 'INTENT (IN) a, b'
    assert parse(Intent, 'intent (in, out) a') == 'INTENT (IN, OUT) a'


def test_entry():
    assert parse(Entry, 'entry a') == 'ENTRY a'
    assert parse(Entry, 'entry a()') == 'ENTRY a'
    assert parse(Entry, 'entry a(b)') == 'ENTRY a (b)'
    assert parse(Entry, 'entry a(b,*)') == 'ENTRY a (b, *)'
    assert parse(Entry, 'entry a bind(c , name="a b")') == \
        'ENTRY a BIND (C, NAME = "a b")'
    assert parse(Entry, 'entry a result (b)') == 'ENTRY a RESULT (b)'
    assert parse(Entry, 'entry a bind(d) result (b)') == \
        'ENTRY a RESULT (b) BIND (D)'
    assert parse(Entry, 'entry a result (b) bind( c )') == \
        'ENTRY a RESULT (b) BIND (C)'
    assert parse(Entry, 'entry a(b,*) result (g)') == \
        'ENTRY a (b, *) RESULT (g)'


def test_import():
    assert parse(Import, 'import') == 'IMPORT'
    assert parse(Import, 'import a') == 'IMPORT a'
    assert parse(Import, 'import::a') == 'IMPORT a'
    assert parse(Import, 'import a , b') == 'IMPORT a, b'


def test_forall():
    assert parse(ForallStmt,
                 'forall (i = 1:n(k,:) : 2) a(i) = i*i*b(i)') == \
        'FORALL (i = 1 : n(k,:) : 2) a(i) = i*i*b(i)'
    assert parse(ForallStmt, 'forall (i=1:n,j=2:3) a(i) = b(i,i)') == \
        'FORALL (i = 1 : n, j = 2 : 3) a(i) = b(i,i)'
    assert parse(ForallStmt,
                 'forall (i=1:n,j=2:3, 1+a(1,2)) a(i) = b(i,i)') == \
        'FORALL (i = 1 : n, j = 2 : 3, 1+a(1,2)) a(i) = b(i,i)'


def test_specificbinding():
    assert parse(SpecificBinding, 'procedure a') == 'PROCEDURE a'
    assert parse(SpecificBinding, 'procedure :: a') == 'PROCEDURE a'
    assert parse(SpecificBinding, 'procedure , NOPASS :: a') == \
        'PROCEDURE , NOPASS :: a'
    assert parse(SpecificBinding, 'procedure , public, pass(x ) :: a') == \
        'PROCEDURE , PUBLIC, PASS (x) :: a'
    assert parse(SpecificBinding, 'procedure(n) a') == 'PROCEDURE (n) a'
    assert parse(SpecificBinding, 'procedure(n), pass :: a') == \
        'PROCEDURE (n) , PASS :: a'
    assert parse(SpecificBinding, 'procedure(n) :: a') == \
        'PROCEDURE (n) a'
    assert parse(SpecificBinding, 'procedure a= >b') == 'PROCEDURE a => b'
    assert parse(SpecificBinding, 'procedure(n),pass :: a =>c') == \
        'PROCEDURE (n) , PASS :: a => c'


def test_genericbinding():
    assert parse(GenericBinding, 'generic :: a=>b') == 'GENERIC :: a => b'
    assert parse(GenericBinding, 'generic, public :: a=>b') == \
        'GENERIC, PUBLIC :: a => b'
    assert parse(GenericBinding, 'generic, public :: a(1,2)=>b ,c') == \
        'GENERIC, PUBLIC :: a(1,2) => b, c'


def test_finalbinding():
    assert parse(FinalBinding, 'final a') == 'FINAL a'
    assert parse(FinalBinding, 'final::a') == 'FINAL a'
    assert parse(FinalBinding, 'final a , b') == 'FINAL a, b'


def test_allocatable():
    assert parse(Allocatable, 'allocatable a') == 'ALLOCATABLE a'
    assert parse(Allocatable, 'allocatable :: a') == 'ALLOCATABLE a'
    assert parse(Allocatable, 'allocatable a (1,2)') == \
        'ALLOCATABLE a (1,2)'
    assert parse(Allocatable, 'allocatable a (1,2) ,b') == \
        'ALLOCATABLE a (1,2), b'


def test_asynchronous():
    assert parse(Asynchronous, 'asynchronous a') == 'ASYNCHRONOUS a'
    assert parse(Asynchronous, 'asynchronous::a') == 'ASYNCHRONOUS a'
    assert parse(Asynchronous, 'asynchronous a , b') == \
        'ASYNCHRONOUS a, b'


def test_bind():
    assert parse(Bind, 'bind(c) a') == 'BIND (C) a'
    assert parse(Bind, 'bind(c) :: a') == 'BIND (C) a'
    assert parse(Bind, 'bind(c) a ,b') == 'BIND (C) a, b'
    assert parse(Bind, 'bind(c) /a/') == 'BIND (C) / a /'
    assert parse(Bind, 'bind(c) /a/ ,b') == 'BIND (C) / a /, b'
    assert parse(Bind, 'bind(c,name="hey") a') == \
        'BIND (C, NAME = "hey") a'


def test_else():
    assert parse(Else, 'else') == 'ELSE'
    assert parse(ElseIf, 'else if (a) then') == 'ELSE IF (a) THEN'
    assert parse(ElseIf, 'else if (a.eq.b(1,2)) then') == \
        'ELSE IF (a.eq.b(1,2)) THEN'


def test_where():
    assert parse(WhereStmt, 'where (1) a=1') == 'WHERE ( 1 ) a = 1'
    assert parse(WhereStmt, 'where (a(1,2)) a=1') == \
        'WHERE ( a(1,2) ) a = 1'


def test_elsewhere():
    assert parse(ElseWhere, 'else where') == 'ELSE WHERE'
    assert parse(ElseWhere, 'elsewhere (1)') == 'ELSE WHERE ( 1 )'
    assert parse(ElseWhere, 'elsewhere(a(1,2))') == \
        'ELSE WHERE ( a(1,2) )'


def test_enumerator():
    assert parse(Enumerator, 'enumerator a') == 'ENUMERATOR a'
    assert parse(Enumerator, 'enumerator:: a') == 'ENUMERATOR a'
    assert parse(Enumerator, 'enumerator a,b') == 'ENUMERATOR a, b'
    assert parse(Enumerator, 'enumerator a=1') == 'ENUMERATOR a=1'
    assert parse(Enumerator, 'enumerator a=1 , b=c(1,2)') == \
        'ENUMERATOR a=1, b=c(1,2)'


def test_fortranname():
    assert parse(FortranName, 'fortranname a') == 'FORTRANNAME a'


def test_threadsafe():
    assert parse(Threadsafe, 'threadsafe') == 'THREADSAFE'


def test_depend():
    assert parse(Depend, 'depend( a) b') == 'DEPEND ( a ) b'
    assert parse(Depend, 'depend( a) ::b') == 'DEPEND ( a ) b'
    assert parse(Depend, 'depend( a,c) b,e') == 'DEPEND ( a, c ) b, e'


def test_check():
    assert parse(Check, 'check(1) a') == 'CHECK ( 1 ) a'
    assert parse(Check, 'check(1) :: a') == 'CHECK ( 1 ) a'
    assert parse(Check, 'check(b(1,2)) a') == 'CHECK ( b(1,2) ) a'
    assert parse(Check, 'check(a>1) :: a') == 'CHECK ( a>1 ) a'


def test_callstatement():
    assert parse(CallStatement, 'callstatement (*func)()', isstrict=1) == \
        'CALLSTATEMENT (*func)()'
    assert parse(CallStatement, 'callstatement i=1;(*func)()',
                 isstrict=1) == \
        'CALLSTATEMENT i=1;(*func)()'


def test_callprotoargument():
    assert parse(CallProtoArgument, 'callprotoargument int(*), double') == \
        'CALLPROTOARGUMENT int(*), double'


def test_pause():
    assert parse(Pause, 'pause') == 'PAUSE'
    assert parse(Pause, 'pause 1') == 'PAUSE 1'
    assert parse(Pause, 'pause "hey"') == 'PAUSE "hey"'
    assert parse(Pause, 'pause "hey pa"') == 'PAUSE "hey pa"'


def test_byte():
    '''Tests various declarations of byte variables.'''
    assert parse(Byte, 'byte') == 'BYTE'
    assert parse(Byte, 'byte a') == 'BYTE a'
    assert parse(Byte, 'byte, a') == 'BYTE a'
    assert parse(Byte, 'byte a ,b') == 'BYTE a, b'
    assert parse(Byte, 'byte :: a ,b') == 'BYTE a, b'
    assert parse(Byte, 'byte a(1,2)') == 'BYTE a(1,2)'
    assert parse(Byte, 'byte :: a(1,2),b') == \
        'BYTE a(1,2), b'
    assert parse(Byte, 'byte external :: a') == \
        'BYTE, external :: a'
    assert parse(Byte, 'byte, external :: a') == \
        'BYTE, external :: a'
    assert parse(Byte, 'byte external , intent(in) :: a') == \
        'BYTE, external, intent(in) :: a'


def test_integer():
    '''Tests various declarations of integer variables.'''
    assert parse(Integer, 'integer') == 'INTEGER'
    assert parse(Integer, 'integer*4') == 'INTEGER*4'
    assert parse(Integer, 'integer*4 a') == 'INTEGER*4 a'
    assert parse(Integer, 'integer*4, a') == 'INTEGER*4 a'
    assert parse(Integer, 'integer*4 a ,b') == 'INTEGER*4 a, b'
    assert parse(Integer, 'integer*4 :: a ,b') == 'INTEGER*4 a, b'
    assert parse(Integer, 'integer*4 a(1,2)') == 'INTEGER*4 a(1,2)'
    assert parse(Integer, 'integer*4 a(2*(1+2)-1)') == 'INTEGER*4 a(2*(1+2)-1)'
    assert parse(Integer, 'integer*4 :: a(1,2),b') == \
        'INTEGER*4 a(1,2), b'
    assert parse(Integer, 'integer*4 external :: a') == \
        'INTEGER*4, external :: a'
    assert parse(Integer, 'integer*4, external :: a') == \
        'INTEGER*4, external :: a'
    assert parse(Integer, 'integer*4 external , intent(in) :: a') == \
        'INTEGER*4, external, intent(in) :: a'
    assert parse(Integer, 'integer(kind=4)') == 'INTEGER(KIND=4)'
    assert parse(Integer, 'integer ( kind = 4)') == 'INTEGER(KIND=4)'
    assert parse(Integer, 'integer(kind=2+2)') == 'INTEGER(KIND=2+2)'
    assert parse(Integer, 'integer(kind=f(4,5))') == \
        'INTEGER(KIND=f(4,5))'
    assert parse(Integer, 'integer a = 5') == 'INTEGER :: a = 5'


def test_const_array_decl():
    '''Tests declarations that set an initial value to an array.'''
    assert parse(Integer, 'integer a(1) = (/1/)') == \
        'INTEGER :: a(1) = (/1/)'
    assert parse(Real, 'real a(3) = (/1.1, 2.2, 3.3/)') == \
        'REAL :: a(3) = (/1.1, 2.2, 3.3/)'
    assert parse(Logical, 'logical :: a(3) = (/true, false, true/)') == \
        'LOGICAL :: a(3) = (/true, false, true/)'
    assert parse(Integer, 'integer a(1) = [1]') == \
        'INTEGER :: a(1) = [1]'
    assert parse(Integer, 'integer a(3) = [1, 2, 3]') == \
        'INTEGER :: a(3) = [1, 2, 3]'


def test_logical():
    '''Tests various declarations of logical variables.'''
    assert parse(Logical, 'logical') == 'LOGICAL'
    assert parse(Logical, 'logical*4') == 'LOGICAL*4'
    assert parse(Logical, 'logical*4 a') == 'LOGICAL*4 a'
    assert parse(Logical, 'logical*4, a') == 'LOGICAL*4 a'
    assert parse(Logical, 'logical*4 a ,b') == 'LOGICAL*4 a, b'
    assert parse(Logical, 'logical*4 :: a ,b') == 'LOGICAL*4 a, b'
    assert parse(Logical, 'logical*4 a(1,2)') == 'LOGICAL*4 a(1,2)'
    assert parse(Logical, 'logical*4 :: a(1,2),b') == \
        'LOGICAL*4 a(1,2), b'
    assert parse(Logical, 'logical*4 external :: a') == \
        'LOGICAL*4, external :: a'
    assert parse(Logical, 'logical*4, external :: a') == \
        'LOGICAL*4, external :: a'
    assert parse(Logical, 'logical*4 external , intent(in) :: a') == \
        'LOGICAL*4, external, intent(in) :: a'
    assert parse(Logical, 'logical(kind=4)') == 'LOGICAL(KIND=4)'
    assert parse(Logical, 'logical ( kind = 4)') == 'LOGICAL(KIND=4)'
    assert parse(Logical, 'logical(kind=2+2)') == 'LOGICAL(KIND=2+2)'
    assert parse(Logical, 'logical(kind=f(4,5))') == \
        'LOGICAL(KIND=f(4,5))'


def test_real():
    '''Test various declarations of real variables. '''

    assert parse(Real, 'real') == 'REAL'
    assert parse(Real, 'real*4') == 'REAL*4'
    assert parse(Real, 'real*4 a') == 'REAL*4 a'
    assert parse(Real, 'real*4, a') == 'REAL*4 a'
    assert parse(Real, 'real*4 a ,b') == 'REAL*4 a, b'
    assert parse(Real, 'real*4 :: a ,b') == 'REAL*4 a, b'
    assert parse(Real, 'real*4 a(1,2)') == 'REAL*4 a(1,2)'
    assert parse(Real, 'real*4 :: a(1,2),b') == \
        'REAL*4 a(1,2), b'
    assert parse(Real, 'real*4 external :: a') == \
        'REAL*4, external :: a'
    assert parse(Real, 'real*4, external :: a') == \
        'REAL*4, external :: a'
    assert parse(Real, 'real*4 external , intent(in) :: a') == \
        'REAL*4, external, intent(in) :: a'
    assert parse(Real, 'real(kind=4)') == 'REAL(KIND=4)'
    assert parse(Real, 'real ( kind = 4)') == 'REAL(KIND=4)'
    assert parse(Real, 'real(kind=2+2)') == 'REAL(KIND=2+2)'
    assert parse(Real, 'real(kind=f(4,5))') == \
        'REAL(KIND=f(4,5))'


def test_double_precision():
    '''Test various declarations of double precision variables. '''

    assert parse(DoublePrecision, 'doubleprecision') == 'DOUBLEPRECISION'
    assert parse(DoublePrecision, 'double precision') == 'DOUBLEPRECISION'
    assert parse(DoublePrecision, 'doubleprecision a') == \
        'DOUBLEPRECISION a'
    assert parse(DoublePrecision, 'double precision, a') == \
        'DOUBLEPRECISION a'
    assert parse(DoublePrecision, 'doubleprecision a ,b') == \
        'DOUBLEPRECISION a, b'
    assert parse(DoublePrecision, 'double precision :: a ,b') == \
        'DOUBLEPRECISION a, b'
    assert parse(DoublePrecision, 'doubleprecision a(1,2)') == \
        'DOUBLEPRECISION a(1,2)'
    assert parse(DoublePrecision, 'double precision :: a(1,2),b') == \
        'DOUBLEPRECISION a(1,2), b'
    assert parse(DoublePrecision, 'doubleprecision external :: a') == \
        'DOUBLEPRECISION, external :: a'
    assert parse(DoublePrecision, 'double precision, external :: a') == \
        'DOUBLEPRECISION, external :: a'
    assert parse(DoublePrecision,
                 'double precision external , intent(in) :: a') == \
        'DOUBLEPRECISION, external, intent(in) :: a'


def test_complex():
    '''Test various declarations of complex variables. '''

    assert parse(Complex, 'complex') == 'COMPLEX'
    assert parse(Complex, 'complex') == 'COMPLEX'
    assert parse(Complex, 'complex*8 a') == 'COMPLEX*8 a'
    assert parse(Complex, 'complex*16, a') == 'COMPLEX*16 a'
    assert parse(Complex, 'complex a ,b') == 'COMPLEX a, b'
    assert parse(Complex, 'complex :: a ,b') == 'COMPLEX a, b'
    assert parse(Complex, 'complex a(1,2)') == 'COMPLEX a(1,2)'
    assert parse(Complex, 'complex :: a(1,2),b') == \
        'COMPLEX a(1,2), b'
    assert parse(Complex, 'complex external :: a') == \
        'COMPLEX, external :: a'
    assert parse(Complex, 'complex, external :: a') == \
        'COMPLEX, external :: a'
    assert parse(Complex, 'complex external , intent(in) :: a') == \
        'COMPLEX, external, intent(in) :: a'
    assert parse(Complex, 'complex(kind=8)') == 'COMPLEX(KIND=8)'
    assert parse(Complex, 'complex ( kind = 16)') == 'COMPLEX(KIND=16)'
    assert parse(Complex, 'complex(kind=4+4)') == 'COMPLEX(KIND=4+4)'
    assert parse(Complex, 'complex(kind=f(4,5))') == \
        'COMPLEX(KIND=f(4,5))'


def test_double_complex():
    '''Test various declarations of double complex variables. '''

    assert parse(DoubleComplex, 'double complex') == 'DOUBLECOMPLEX'
    assert parse(DoubleComplex, 'double complex') == 'DOUBLECOMPLEX'
    assert parse(DoubleComplex, 'double complex a') == 'DOUBLECOMPLEX a'
    assert parse(DoubleComplex, 'double complex, a') == 'DOUBLECOMPLEX a'
    assert parse(DoubleComplex, 'double complex a ,b') == \
        'DOUBLECOMPLEX a, b'
    assert parse(DoubleComplex, 'double complex :: a ,b') == \
        'DOUBLECOMPLEX a, b'
    assert parse(DoubleComplex, 'double complex a(1,2)') == \
        'DOUBLECOMPLEX a(1,2)'
    assert parse(DoubleComplex, 'double complex :: a(1,2),b') == \
        'DOUBLECOMPLEX a(1,2), b'
    assert parse(DoubleComplex, 'double complex external :: a') == \
        'DOUBLECOMPLEX, external :: a'
    assert parse(DoubleComplex, 'double complex, external :: a') == \
        'DOUBLECOMPLEX, external :: a'
    assert parse(DoubleComplex,
                 'double complex external , intent(in) :: a') == \
        'DOUBLECOMPLEX, external, intent(in) :: a'
    assert parse(DoubleComplex, 'double complex(kind=4)') == \
        'DOUBLECOMPLEX(KIND=4)'
    assert parse(DoubleComplex, 'double complex ( kind = 4)') == \
        'DOUBLECOMPLEX(KIND=4)'
    assert parse(DoubleComplex, 'double complex(kind=2+2)') == \
        'DOUBLECOMPLEX(KIND=2+2)'
    assert parse(DoubleComplex, 'double complex(kind=f(4,5))') == \
        'DOUBLECOMPLEX(KIND=f(4,5))'


def test_character():
    assert parse(Character, 'character') == 'CHARACTER'
    assert parse(Character, 'character*2') == 'CHARACTER(LEN=2)'
    assert parse(Character, 'character**') == 'CHARACTER(LEN=*)'
    assert parse(Character, 'character*(2)') == 'CHARACTER(LEN=2)'
    assert parse(Character, 'character*(len =2)') == 'CHARACTER(LEN=2)'
    assert parse(Character, 'character*(len =2)') == 'CHARACTER(LEN=2)'
    assert parse(Character, 'character*(len =:)') == 'CHARACTER(LEN=:)'
    assert parse(Character, 'character(len =2)') == 'CHARACTER(LEN=2)'
    assert parse(Character, 'character(2)') == 'CHARACTER(LEN=2)'
    assert parse(Character, 'character(kind=2)') == 'CHARACTER(KIND=2)'
    assert parse(Character, 'character(kind=2,len=3)') == \
        'CHARACTER(LEN=3, KIND=2)'
    assert parse(Character, 'character(lEN=3,kind=2)') == \
        'CHARACTER(LEN=3, KIND=2)'
    assert parse(Character, 'character(len=3,kind=2)', isstrict=True) == \
        'CHARACTER(LEN=3, KIND=2)'
    assert parse(Character, 'chaRACTER(len=3,kind=fA(1,2))',
                 isstrict=True) == \
        'CHARACTER(LEN=3, KIND=fA(1,2))'
    assert parse(Character, 'character(len=3,kind=fA(1,2))') == \
        'CHARACTER(LEN=3, KIND=fa(1,2))'


def test_implicit():
    assert parse(Implicit, 'implicit none') == 'IMPLICIT NONE'
    assert parse(Implicit, 'implicit') == 'IMPLICIT NONE'
    assert parse(Implicit, 'implicit integer (i-m)') == \
        'IMPLICIT INTEGER ( i-m )'
    assert parse(Implicit, 'implicit integer (i-m,p,q-r)') == \
        'IMPLICIT INTEGER ( i-m, p, q-r )'
    assert parse(Implicit, 'implicit integer (i-m), real (z)') ==\
        'IMPLICIT INTEGER ( i-m ), REAL ( z )'


def test_type_bound_array_access():
    ''' Check that we can parse code that calls a type-bound procedure
    on the element of an array (of derived types) '''
    parsed_code = parse(Call, 'call an_array(idx)%a_func(arg)')
    assert parsed_code == "CALL an_array(idx)%a_func(arg)"
    # Routine being called has no arguments - it is valid Fortran to
    # omit the parentheses entirely in this case
    parsed_code = parse(Call, 'call an_array(idx)%a_func()')
    assert parsed_code == "CALL an_array(idx)%a_func"
    parsed_code = parse(Call, 'call an_array(idx)%a_func')
    assert parsed_code == "CALL an_array(idx)%a_func"
    # Perversely put in a character string arg containing '('
    parsed_code = parse(Call, 'call an_array(idx)%a_func("(")')
    assert parsed_code == 'CALL an_array(idx)%a_func("(")'
    # Pass an array element as an argument
    parsed_code = parse(Call, 'call an_array(idx)%a_func(b(3))')
    assert parsed_code == 'CALL an_array(idx)%a_func(b(3))'


def test_invalid_type_bound_array_access():  # pylint: disable=invalid-name
    ''' Check that a call to a type-bound procedure with incorrect
    syntax is flagged as invalid '''
    with pytest.raises(ValueError) as excinfo:
        _ = parse(Call, 'call an_array(idx)%a_func)')
    assert "with Call pattern failed" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = parse(Call, 'call an_array(idx)%)')
    assert "with Call pattern failed" in str(excinfo.value)


def test_analyze_errors():
    ''' Tests that AnalyzeErrors are raised as expected. It also tests
    for various calling-sequence issues, e.g. parsing or analyzing twice,
    or calling analyze() without calling parse() first.'''
    from fparser import api
    source_str = """none subroutine test()
      end
    """
    with pytest.raises(AnalyzeError) as error:
        _ = api.parse(source_str, isfree=True, isstrict=False)
    assert "no parse pattern found" in str(error.value)

    source_str = """subroutine test()
        incorrect :: c
      end
    """
    with pytest.raises(AnalyzeError) as error:
        _ = api.parse(source_str, isfree=False, isstrict=False)
    assert "no parse pattern found" in str(error.value)

    source_str = """subroutine test()
      end
    """
    from fparser.one.parsefortran import FortranParser
    reader = api.get_reader(source_str)
    parser = FortranParser(reader)

    # Handle analyze before parsing (does not raise an exception atm)
    parser.analyze()
    # Cover case that parsing is called twice
    parser.parse()
    parser.parse()

    # Check if analyse is called twice
    parser.analyze()
    parser.analyze()



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 : This file tests the support for a format item
list. List classes are usually generated, and therefore are not
directly tested, but in this case we need to make a specific list
class due to the complexity of supporting Hollerith format items.

Note, the hollerith format was deprecated in Fortran77 and removed in
Fortran95. However, Fortran compilers still support it.

'''

import pytest
from fparser.two.Fortran2003 import Format_Item_List
from fparser.two.utils import NoMatchError


def test_non_hollerith(f2003_create):
    '''Check that some basic format specifications are parsed
    correctly. Individual specifications are tested by the associated
    classes.

    '''
    myinput = "E2.2, 'HELLO'"
    ast = Format_Item_List(myinput)
    assert myinput in str(ast)
    assert repr(ast).replace("u", "") == (
        "Format_Item_List(',', (Format_Item(None, Data_Edit_Desc_C1002('E', "
        "Digit_String('2', None), Int_Literal_Constant('2', None), None)), "
        "Char_Literal_Constant(\"'HELLO'\", None)))")


def test_mixed_hollerith_1(f2003_create, monkeypatch):
    '''Check that a hollerith item is parsed correctly. We do this as a
    hollerith item can contain commas so we can't use a simple
    format-item-list class to split items.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    myinput = "2Hab, 'HELLO'"
    ast = Format_Item_List(myinput)
    assert myinput in str(ast)
    assert repr(ast).replace("u", "") == (
        "Format_Item_List(',', (Hollerith_Item('ab'), "
        "Char_Literal_Constant(\"'HELLO'\", None)))")


def test_mixed_hollerith_2(f2003_create, monkeypatch):
    '''Check that a hollerith item is parsed correctly. We do this as a
    hollerith item can contain commas so we can't use a simple
    format-item-list class to split items.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    myinput = "'HELLO', 2hab"
    ast = Format_Item_List(myinput)
    assert myinput.lower() in str(ast).lower()
    assert repr(ast).replace("u", "") == (
        "Format_Item_List(',', (Char_Literal_Constant(\"'HELLO'\", None), "
        "Hollerith_Item('ab')))")


def test_hollerith_only(f2003_create, monkeypatch):
    '''Check that a hollerith item is parsed correctly. We do this as a
    hollerith item can contain commas so we can't use a simple
    format-item-list class to split items.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    myinput = "3Habc,2hab"
    ast = Format_Item_List(myinput)
    assert str(ast) == "3Habc, 2Hab"
    assert repr(ast) == (
        "Format_Item_List(',', (Hollerith_Item('abc'), "
        "Hollerith_Item('ab')))")


def test_hollerith_only_spaces(f2003_create, monkeypatch):
    '''Check that a hollerith item is parsed correctly in the presence of
    additional spaces. We do this as a hollerith item must have the
    correct number of characters so any spaces that are not required
    must be removed.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    myinput = "  3Habc  ,  2hab  "
    ast = Format_Item_List(myinput)
    assert str(ast) == "3Habc, 2Hab"
    assert repr(ast) == (
        "Format_Item_List(',', (Hollerith_Item('abc'), "
        "Hollerith_Item('ab')))")


def test_errors(f2003_create, monkeypatch):
    '''test some list errors. Individual item errors will be picked up by
    the subclasses.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    for myinput in [None, "", "  ", "E2.2  2Hab", "E2.2, E2.2 E2.2",
                    "2Hab,2Ha,2Hab", "2Hab,2Hab x,2Hab", "2Hab,2Hab,2Ha",
                    "2Hab,2Hab,2Hab x"]:
        with pytest.raises(NoMatchError):
            _ = Format_Item_List(myinput)



# Copyright (c) 2017-2020 Science and Technology Facilities Council
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Module containing tests for aspects of fparser2 related to comments '''

import pytest
from fparser.two.Fortran2003 import Program, Comment, Subroutine_Subprogram
from fparser.two.utils import walk
from fparser.api import get_reader

from fparser.two.parser import ParserFactory
# this is required to setup the fortran2003 classes
_ = ParserFactory().create(std="f2003")


def test_ignore_comments():
    ''' Check that the parser does throw away comments when requested '''
    tree = Program(get_reader('''\
PROGRAM a_prog
! A full line comment
PRINT *, "Hello" ! This block gets executed
END PROGRAM a_prog
    ''', ignore_comments=True))
    gen = str(tree)
    assert ("PROGRAM a_prog\n"
            "  PRINT *, \"Hello\"\n"
            "END PROGRAM a_prog" in gen)
    assert "full line comment" not in gen
    assert "block gets executed" not in gen

    # Check that the default behaviour is to ignore comments
    tree = Program(get_reader('''\
PROGRAM a_prog
! A full line comment
PRINT *, "Hello" ! This block gets executed
END PROGRAM a_prog
    '''))
    gen = str(tree)
    assert "full line comment" not in gen
    assert "block gets executed" not in gen


def test_simple_prog():
    ''' Tests simplest case of comments in a program unit '''
    tree = Program(get_reader('''\
PROGRAM a_prog
! A full line comment
PRINT *, "Hello" ! This block gets executed
END PROGRAM a_prog
    ''', ignore_comments=False))
    assert (str(tree) == "PROGRAM a_prog\n"
            "  ! A full line comment\n"
            "  PRINT *, \"Hello\"\n"
            "  ! This block gets executed\n"
            "END PROGRAM a_prog\n")


def test_ifthen():
    ''' Tests for comments within an if-then block '''
    cls = Program
    tree = cls(get_reader('''\
PROGRAM a_prog
IF(.TRUE.)THEN
! A full line comment
PRINT *, "Hello"
! Another full line comment
END IF
END PROGRAM a_prog
    ''', ignore_comments=False))
    assert (str(tree) == "PROGRAM a_prog\n"
            "  IF (.TRUE.) THEN\n"
            "    ! A full line comment\n"
            "    PRINT *, \"Hello\"\n"
            "    ! Another full line comment\n"
            "  END IF\n"
            "END PROGRAM a_prog\n")


def test_inline_ifthen():
    ''' Tests for in-line comments within an if-then block '''
    tree = Program(get_reader('''\
PROGRAM a_prog
IF(.TRUE.)THEN
PRINT *, "Hello" ! An in-line comment here
END IF
! A comment after a block
END PROGRAM a_prog
    ''', ignore_comments=False))
    assert isinstance(tree, Program)
    assert (str(tree) == "PROGRAM a_prog\n"
            "  IF (.TRUE.) THEN\n"
            "    PRINT *, \"Hello\"\n"
            "    ! An in-line comment here\n"
            "  END IF\n"
            "  ! A comment after a block\n"
            "END PROGRAM a_prog\n")


@pytest.mark.xfail(reason="fails to preserve formatting in fixed format")
def test_fixed_fmt():
    ''' Test that we handle comments in fixed-format mode '''
    reader = get_reader('''\
      subroutine foo
c this is a subroutine
        end subroutine foo''', isfree=False, ignore_comments=True)
    cls = Subroutine_Subprogram
    obj = cls(reader)
    assert isinstance(obj, cls), repr(obj)
    assert str(obj) == 'SUBROUTINE foo\nEND SUBROUTINE foo'
    assert (repr(obj) == "Subroutine_Subprogram(Subroutine_Stmt(None, "
            "Name('foo'), None, None), End_Subroutine_Stmt('SUBROUTINE', "
            "Name('foo')))")
    reader = get_reader('''\
      subroutine foo
c this is a subroutine
      end subroutine foo''', isfree=False, ignore_comments=False)
    cls = Subroutine_Subprogram
    obj = cls(reader)
    fort = obj.tofortran(isfix=True)
    assert isinstance(obj, cls), repr(obj)
    assert (fort == '      SUBROUTINE foo\nc this is a subroutine\n'
            '      END SUBROUTINE foo')
    assert (repr(obj) == "Subroutine_Subprogram(Subroutine_Stmt(None, "
            "Name('foo'), None, None), End_Subroutine_Stmt('SUBROUTINE', "
            "Name('foo')))")


@pytest.mark.xfail(reason="fails to preserve formatting in fixed format")
def test_fixed_continuation():
    ''' Check that we handle comments that occur within fixed-format
    continued lines '''
    reader = get_reader('''\
      subroutine foo
c this is a subroutine
      i = 1 +
c this comment is a problem
     &2 + 3
      end subroutine foo''', isfree=False, ignore_comments=False)
    cls = Subroutine_Subprogram
    obj = cls(reader)
    assert isinstance(obj, cls)
    assert "i = 1 + 2 + 3\n     c this comment is a problem" in str(obj)


def test_prog_comments():
    ''' Unit tests for lines in programs containing comments '''
    cls = Program
    reader = get_reader('''\
   ! A troublesome comment
   program foo
     ! A full comment line
     write(*,*) my_int ! An in-line comment
    end program foo
! A really problematic comment
''', isfree=True, ignore_comments=False)

    obj = cls(reader)
    assert type(obj) == Program
    # Check that the AST has the expected structure:
    # Program
    #   |--> Comment
    #   |--> Main_Program
    #   .    |--> Program_Stmt
    #   .    |--> Specification_Part
    #   .    .    \--> Implicit_Part
    #   .    .         \--> Comment
    #        |--> Execution_Part
    #        |    |--> Write_Stmt
    #        |    \--> Comment
    #   .    .
    #   .
    #   |--> Comment
    from fparser.two.Fortran2003 import Main_Program, Write_Stmt, \
        End_Program_Stmt
    walk(obj.children, Comment, debug=True)
    assert type(obj.content[0]) == Comment
    assert str(obj.content[0]) == "! A troublesome comment"
    assert type(obj.content[1]) == Main_Program
    main_prog = obj.content[1]
    assert type(main_prog.content[1].content[0].content[0]) == Comment
    assert (str(main_prog.content[1].content[0].content[0]) ==
            "! A full comment line")
    exec_part = main_prog.content[2]
    assert type(exec_part.content[0]) == Write_Stmt
    # Check that we have the in-line comment as a second statement
    assert len(exec_part.content) == 2
    assert type(exec_part.content[1]) == Comment
    assert type(main_prog.content[3]) == End_Program_Stmt
    assert "! An in-line comment" in str(obj)
    # Check that we still have the ending comment
    assert type(obj.content[-1]) == Comment
    assert str(obj).endswith("! A really problematic comment")


def test_module_comments():
    ''' Tests for comments in modules '''
    source = '''! This is a module
      module my_mod
        implicit none
        private
      end module my_mod
'''
    # Test when the reader is explicitly set to free-form mode
    reader = get_reader(source, isfree=True, ignore_comments=False)
    prog_unit = Program(reader)
    assert type(prog_unit.content[0]) == Comment
    assert str(prog_unit.content[0]) == "! This is a module"


def test_function_comments():
    ''' Tests for comments in functions '''
    source = '''\
function my_mod()
  ! This is a function
  implicit none
  integer my_mod
  ! Comment1
  my_mod = 1 ! Comment2
  ! That was a function
end function my_mod
'''
    from fparser.two.Fortran2003 import Function_Subprogram
    reader = get_reader(source, isfree=True, ignore_comments=False)
    fn_unit = Function_Subprogram(reader)
    # <class 'fparser.two.Fortran2003.Function_Stmt'>
    #    <type 'NoneType'>
    #    <class 'fparser.two.Fortran2003.Name'>
    #    <type 'NoneType'>
    #    <type 'NoneType'>
    # <class 'fparser.two.Fortran2003.Specification_Part'>
    #   <class 'fparser.two.Fortran2003.Implicit_Part'>
    #     <class 'fparser.two.Fortran2003.Comment'>
    #       <type 'str'>, "'! This is a function'"
    comment = fn_unit.content[1].content[0].content[0]
    assert isinstance(comment, Comment)
    assert "! This is a function" in str(comment)
    comment = fn_unit.content[1].content[2].content[0]
    assert isinstance(comment, Comment)
    assert "! Comment1" in str(comment)
    exec_part = fn_unit.content[2]
    comment = exec_part.content[1]
    assert isinstance(comment, Comment)
    assert "! Comment2" in str(comment)
    comment = exec_part.content[2]
    assert isinstance(comment, Comment)
    assert "! That was a function" in str(comment)


def test_subroutine_comments():
    ''' Tests for comments in subroutines '''
    source = '''\
subroutine my_mod()
  ! First comment
  implicit none
  integer my_mod
  ! Body comment
  my_mod = 1 ! Inline comment
  ! Ending comment
end subroutine my_mod
'''
    reader = get_reader(source, isfree=True, ignore_comments=False)
    fn_unit = Subroutine_Subprogram(reader)
    assert isinstance(fn_unit, Subroutine_Subprogram)
    walk(fn_unit.children, Comment, debug=True)
    spec_part = fn_unit.content[1]
    comment = spec_part.content[0].content[0]
    assert isinstance(comment, Comment)
    assert "! First comment" in str(comment)
    comment = spec_part.content[2].content[0]
    assert isinstance(comment, Comment)
    assert comment.parent is spec_part.content[2]
    assert "! Body comment" in str(comment)
    exec_part = fn_unit.content[2]
    comment = exec_part.content[1]
    assert isinstance(comment, Comment)
    assert comment.parent is exec_part
    assert "! Inline comment" in str(comment)


def test_derived_type():
    ''' Test for comments within declaration of derived type '''
    source = '''\
type my_type ! Inline comment1
  ! First comment
  integer :: start ! Inline comment2
  integer :: stop
  ! Ending comment
end type my_type
'''
    from fparser.two.Fortran2003 import Derived_Type_Def
    reader = get_reader(source, isfree=True, ignore_comments=False)
    dtype = Derived_Type_Def(reader)
    assert isinstance(dtype, Derived_Type_Def)
    comments = [(1, "! Inline comment1"),
                (2, "! First comment"),
                (4, "! Inline comment2"),
                (6, "! Ending comment")]
    for comment in comments:
        idx = comment[0]
        assert isinstance(dtype.content[idx], Comment)
        assert comment[1] in str(dtype.content[idx])


def test_action_stmts():
    ''' Tests for comments within action statements such as allocate '''
    # We have to have our allocate() within some other block otherwise
    # we lose the comment (it is deferred until after the current,
    # continuted line is complete but once that happens, the match for
    # the allocate() is done and the reader/parser returns without
    # ever dealing with the comment).
    source = '''\
if(.true.)then
allocate(my_array(size), &
         ! that's a big array
         my_array2(size))
end if
'''
    from fparser.two.Fortran2003 import If_Construct, Allocate_Stmt
    from fparser.two.utils import get_child
    reader = get_reader(source, isfree=True, ignore_comments=False)
    ifstmt = If_Construct(reader)
    assert isinstance(ifstmt, If_Construct)
    assert isinstance(ifstmt.content[1], Allocate_Stmt)
    assert "a big array" in str(ifstmt)
    cmt = get_child(ifstmt, Comment)
    assert cmt.parent is ifstmt



##############################################################################
# Copyright (c) 2017-2020 Science and Technology Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
# Modified M.Hambley, UK Met Office
##############################################################################
'''
Test battery associated with fparser.two.pattern_tools package.
'''

import pytest
import fparser.two.pattern_tools


def test_name_pattern():
    '''
    Tests inherited from implementation source.
    '''
    assert fparser.two.pattern_tools.name.match('a1_a')
    assert abs(fparser.two.pattern_tools.name).match('a1_a')
    assert not abs(fparser.two.pattern_tools.name).match('a1_a[]')


def test_kind_param_pattern():
    '''
    Tests inherited from implementation source.
    '''
    match = abs(fparser.two.pattern_tools.kind_param)
    assert match.match('23')
    assert match.match('SHORT')


def test_signed_digit_pattern():
    '''
    Tests inherited from implementation source.
    '''
    match = abs(fparser.two.pattern_tools.signed_digit_string)
    assert match.match('23')
    assert match.match('+ 23')
    assert match.match('- 23')
    assert match.match('-23')
    assert not match.match('+n')


def test_thing():
    '''
    Tests inherited from implementation source.
    '''
    match = ~fparser.two.pattern_tools.sign.named() \
        + fparser.two.pattern_tools.digit_string.named('number')
    result = match.match('23')
    assert result.groupdict() == {'number': '23', 'sign': None}
    result = match.match('- 23')
    assert result.groupdict() == {'number': '23', 'sign': '-'}


def test_char_literal_const_pattern():
    '''
    Tests inherited from implementation source.
    '''
    match = abs(fparser.two.pattern_tools.char_literal_constant)
    assert match.match('"adadfa"')
    assert match.match('"adadfa""adad"')
    assert match.match('HEY_"adadfa"')
    assert match.match('HEY _ "ad\tadfa"')
    assert not match.match('adadfa')


def test_multi_op_pattern():
    '''
    Tests inherited from implementation source.
    '''
    match = fparser.two.pattern_tools.mult_op.named()
    assert match.rsplit('a *  b')
    assert match.lsplit('a * c* b') == ('a', '*', 'c* b')
    assert match.rsplit('a * c* b') == ('a * c', '*', 'b')
    assert match.lsplit('a * b ** c') == ('a', '*', 'b ** c')
    assert match.rsplit('a * b ** c') == ('a', '*', 'b ** c')
    assert match.lsplit('a * b ** c * d') == ('a', '*', 'b ** c * d')
    assert match.rsplit('a * b ** c * d') == ('a * b ** c', '*', 'd')


def test_power_op_pattern():
    '''
    Tests inherited from implementation source.
    '''
    match = fparser.two.pattern_tools.power_op.named()
    assert match.rsplit('a **  b')
    assert match.lsplit('a * b ** c') == ('a * b', '**', 'c')
    assert match.rsplit('a * b ** c') == ('a * b', '**', 'c')
    assert match.lsplit('a ** b ** c') == ('a', '**', 'b ** c')
    assert match.rsplit('a ** b ** c') == ('a ** b', '**', 'c')


@pytest.mark.parametrize("pattern",
                         ["ALLOCATABLE", "ASYNCHRONOUS", "EXTERNAL", "INTENT",
                          "INTRINSIC", "OPTIONAL", "PARAMETER", "POINTER",
                          "PROTECTED", "SAVE", "TARGET", "VALUE", "VOLATILE"])
def test_attr_spec_pattern(pattern):
    '''Tests the attr_spec and abs_attr_spec patterns for all valid
    matches. The difference between the two is that the former matches
    even if there is additional content after the matching string, but
    the latter (abs) does not.

    '''
    attr_spec = fparser.two.pattern_tools.attr_spec
    assert attr_spec.match(pattern.upper())
    assert attr_spec.match(pattern.lower())
    assert not attr_spec.match("X"+pattern)
    assert attr_spec.match(pattern+"X")

    abs_attr_spec = fparser.two.pattern_tools.abs_attr_spec
    assert abs_attr_spec.match(pattern.upper())
    assert abs_attr_spec.match(pattern.lower())
    assert not abs_attr_spec.match("X"+pattern)
    assert not abs_attr_spec.match(pattern+"X")



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2008 rule R202 : This file tests the addition of
submodules in Fortran2008 for the Program-unit rule.

'''

import pytest
from fparser.two.utils import NoMatchError
from fparser.api import get_reader
from fparser.two.Fortran2008 import Program_Unit


def test_other(f2008_create):
    '''Test that something other than submodule can still be parsed
    i.e. by adding submodule we've not broken the existing
    program-unit options.

    '''
    reader = get_reader('''\
      subroutine test()
      end subroutine
      ''')
    ast = Program_Unit(reader)
    assert "SUBROUTINE test\n" \
        "END SUBROUTINE" in str(ast)


def test_submodule(f2008_create):
    '''Test that submodule as a top-level program unit can be parsed.'''
    reader = get_reader('''\
      submodule (foobar) bar
      end
      ''')
    ast = Program_Unit(reader)
    assert "SUBMODULE (foobar) bar\n" \
        "END" in str(ast)


def test_submodule_nomatch(f2008_create):
    '''Test an exception is raised if there is a syntax error.'''
    reader = get_reader('''\
      submod (foobar) bar
      end
      ''')
    with pytest.raises(NoMatchError) as excinfo:
        dummy_ = Program_Unit(reader)
    assert ("at line 1\n>>>      submod (foobar) bar\n"
            in str(excinfo.value))



# Copyright (c) 2018-2019 Science and Technology Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test fparser scripts

'''

# pylint: disable=too-few-public-methods

import pytest
from fparser.scripts import fparser2, read

# fparser2.py script function runner()

# Create a dummy class (DummyArgs) with the required attribute to pass
# into runner() as an argument options class in subsequent tests


class DummyArgs(object):
    ''' dummy object pretending to be the argument options '''
    mode = "auto"
    task = "show"
    std = "f2003"


def test_runner_no_files(capsys):
    '''Test that the script deals with no files provided as expected.'''
    # run the relevant script method (runner())
    with pytest.raises(SystemExit) as excinfo:
        fparser2.runner(None, DummyArgs(), [])
    assert str(excinfo.value) == "1"
    # capture the output and check that the appropriate error has been reported
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert "Error: No fortran files specified" in stderr


def test_runner_non_existant_file(capsys):
    '''Test that the script reports an error when the file name that is
    provided does not exist.

    '''
    # run the relevant script method (runner())
    fparser2.runner(None, DummyArgs(), ["idontexist.txt"])
    # capture the output and check that the appropriate error has been reported
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert "File: 'idontexist.txt'" in stderr
    assert "No such file or directory: 'idontexist.txt'" in stderr


def test_runner_no_file_multi(capsys, tmpdir):
    '''Test that the script reports an error when the file name that is
    provided does not exist and continues to parse any subsequent
    file(s).

    '''
    # Create a temporary file containing Fortran code to pass into runner()
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # run the relevant script method (runner())
    fparser2.runner(None, DummyArgs(), ["idontexist.txt", my_file.strpath])
    # capture the output and check that the appropriate error has been
    # reported
    stdout, stderr = capsys.readouterr()
    assert "PROGRAM hello\nEND PROGRAM hello\n" in stdout
    assert "File: 'idontexist.txt'" in stderr
    assert "No such file or directory: 'idontexist.txt'" in stderr
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_runner_set_mode(tmpdir, capsys):
    '''Test that the script can change mode.'''
    # Create a temporary file containing Fortran code to pass into runner()
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Create a dummy class with the required attribute to pass into
    # runner() as an argument options class

    class DummyArgsFree(object):
        ''' dummy object pretending to be the argument options '''
        mode = "free"
        task = "show"
        std = "f2003"
    # run the relevant script method (runner())
    fparser2.runner(None, DummyArgsFree(), [my_file.strpath])
    # capture the output and check that the code has been output
    stdout, stderr = capsys.readouterr()
    assert "PROGRAM hello\nEND PROGRAM hello\n" in stdout
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_runner_syntax_error(tmpdir, capsys):
    '''Test that the script deals with code with an invalid syntax.'''
    # Create a temporary file containing Fortran code to pass into runner()
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("prog error\nend program error\n")
    # run the relevant script method (runner())
    fparser2.runner(None, DummyArgs(), [my_file.strpath])
    # capture the output and check that the appropriate error has been
    # reported
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert "File: '" in stderr and "hello.f90'" in stderr
    assert "Syntax error: at line 1\n>>>prog error" in stderr


def test_runner_internal_error(tmpdir, monkeypatch, capsys):
    '''Test that the script deals with an internal error as
    expected. Provide the same file twice to check that the parser
    continues to parse any additional files.

    '''
    # Create a temporary file containing Fortran code to pass into runner()
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Create a dummy function that replaces the parser

    def dummy_parser(_self, std="f2003"):
        ''' dummy function that simply raises an internal error '''
        raise InternalError(std)
    # monkeypatch the parser so that it returns an InternalError exception.
    from fparser.two.parser import ParserFactory
    from fparser.two.utils import InternalError
    monkeypatch.setattr(ParserFactory, "create", dummy_parser)
    # run the relevant script method (runner())
    fparser2.runner(None, DummyArgs(), [my_file.strpath, my_file.strpath])
    # capture the output and check that the appropriate error has been reported
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert stderr.count("Internal error in fparser: 'f2003'") == 2


def test_runner_output_task_show(tmpdir, capsys):
    '''Test that the script outputs the code it has parsed with the
    'task' option set to "show".

    '''
    class DummyArgsTask(object):
        ''' dummy object pretending to be the argument options '''
        mode = "free"
        task = "show"
        std = "f2003"

    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Run the relevant script method (runner()).
    fparser2.runner(None, DummyArgsTask(), [my_file.strpath])
    # Capture the output and check that the code has been output.
    stdout, stderr = capsys.readouterr()
    assert "PROGRAM hello\nEND PROGRAM hello\n" in stdout
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_runner_output_task_repr(tmpdir, capsys):
    '''Test that the script outputs the repr representation of the code it
    has parsed with the 'task' option set to "repr".

    '''
    class DummyArgsTask(object):
        ''' dummy object pretending to be the argument options '''
        mode = "free"
        task = "repr"
        std = "f2003"

    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Run the relevant script method (runner()).
    fparser2.runner(None, DummyArgsTask(), [my_file.strpath])
    # Capture the output and check that the repr of the code has been
    # produced.
    stdout, stderr = capsys.readouterr()
    assert "File: '" in stderr and "hello.f90'" in stderr
    assert (stdout == "Program(Main_Program(Program_Stmt('PROGRAM',"
            " Name('hello')), End_Program_Stmt('PROGRAM', Name('hello'))))\n")


def test_runner_output_task_none(tmpdir, capsys):
    '''Test that the script outputs nothing when the 'task' option is set
    to "none".

    '''
    class DummyArgsTask(object):
        ''' dummy object pretending to be the argument options '''
        mode = "free"
        task = "none"
        std = "f2003"

    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Run the relevant script method (runner()).
    fparser2.runner(None, DummyArgsTask(), [my_file.strpath])
    # Capture the output and check that nothig has been produced.
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_runner_multi_output(tmpdir, capsys):
    '''Test that the script outputs the code it has parsed when there are
    multiple files specified

    '''
    # Create a temporary file containing Fortran code to pass into runner()
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # run the relevant script method (runner())
    fparser2.runner(None, DummyArgs(), [my_file.strpath, my_file.strpath])
    # capture the output and check that the code has been output
    stdout, stderr = capsys.readouterr()
    assert ("PROGRAM hello\nEND PROGRAM hello\n"
            "PROGRAM hello\nEND PROGRAM hello\n") in stdout
    assert stderr.count("File: '") == 2
    assert stderr.count("hello.f90'") == 2


def test_runner_multi_output_except(tmpdir, capsys):
    '''Test that if the script encounters a file with a syntax error then
    it continues trying to parse subsequent files.

    '''
    # Create a temporary file containing Fortran code to pass into runner()
    my_file1 = tmpdir.mkdir("sub1").join("broken.f90")
    my_file1.write("prog hello\nen\n")
    my_file2 = tmpdir.mkdir("sub2").join("hello.f90")
    my_file2.write("program hello\nend program hello\n")
    # run the relevant script method (runner())
    fparser2.runner(None, DummyArgs(), [my_file1.strpath, my_file2.strpath])
    # capture the output and check that the valid code has been output
    # and the syntax error has been raised.
    stdout, stderr = capsys.readouterr()
    assert stdout == "PROGRAM hello\nEND PROGRAM hello\n"
    assert "Syntax error: at line 1\n>>>prog hello" in stderr
    assert stderr.count("File: '") == 2
    assert "hello.f90'" in stderr
    assert "broken.f90'" in stderr

# fparser2.py script function main()


def test_main_output_task_default(tmpdir, capsys, monkeypatch):
    '''Test that the script main() function outputs the code it has parsed
    by default.'''
    import sys
    # Create a temporary file containing Fortran code to pass into runner()
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Use monkeypatch to spoof the command-line argument
    monkeypatch.setattr(sys, "argv", ["fparser2", my_file.strpath])
    # run the relevant script method (main())
    fparser2.main()
    # capture the output and check that the code has been output
    stdout, stderr = capsys.readouterr()
    assert stdout == "PROGRAM hello\nEND PROGRAM hello\n"
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_main_output_task_show(tmpdir, capsys, monkeypatch):
    '''Test that the script main() function outputs the code it has parsed
    when --task=show.

    '''
    import sys
    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Use monkeypatch to spoof the command-line argument.
    monkeypatch.setattr(sys, "argv", ["fparser2", "--task=show",
                                      my_file.strpath])
    # Run the relevant script method (main()).
    fparser2.main()
    # Capture the output and check that the code has been output.
    stdout, stderr = capsys.readouterr()
    assert stdout == "PROGRAM hello\nEND PROGRAM hello\n"
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_main_output_task_repr(tmpdir, capsys, monkeypatch):
    '''Test that the script main() function outputs the 'repr' of the code
    it has parsed when --task=repr.

    '''
    import sys
    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Use monkeypatch to spoof the command-line argument.
    monkeypatch.setattr(sys, "argv", ["fparser2", "--task=repr",
                                      my_file.strpath])
    # Run the relevant script method (main()).
    fparser2.main()
    # Capture the output and check that the code has been output.
    stdout, stderr = capsys.readouterr()
    assert stdout == ("Program(Main_Program(Program_Stmt('PROGRAM',"
                      " Name('hello')), End_Program_Stmt('PROGRAM',"
                      " Name('hello'))))\n")
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_main_output_task_none(tmpdir, capsys, monkeypatch):
    '''Test that the script main() function outputs nothing when
    --task=none.

    '''
    import sys
    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Use monkeypatch to spoof the command-line argument.
    monkeypatch.setattr(sys, "argv", ["fparser2", "--task=none",
                                      my_file.strpath])
    # Run the relevant script method (main()).
    fparser2.main()
    # Capture the output and check that the code has been output.
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert "File: '" in stderr and "hello.f90'" in stderr


def test_main_output_task_invalid(tmpdir, capsys, monkeypatch):
    '''Test that the script main() function prints an error when an
    invalid task option is provided.

    '''
    import sys
    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Use monkeypatch to spoof the command-line argument.
    monkeypatch.setattr(sys, "argv", ["fparser2", "--task=invalid",
                                      my_file.strpath])
    with pytest.raises(SystemExit):
        # Run the relevant script method (main()).
        fparser2.main()
    # Capture the output and check that the appropriate error has been
    # written.
    stdout, stderr = capsys.readouterr()
    assert ("fparser2: error: option --task: invalid choice: 'invalid'"
            "" in stderr)
    assert stdout == ""


def test_main_output_std_f2003(capsys, tmpdir, monkeypatch):
    '''Test that the script main() function parses using the Fortran2003
    standard when --std=f2003. This test attempts to use submodules
    which are not supported in Fortran2003 so should give a syntax
    error.

    '''
    import sys
    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("submodule.f90")
    my_file.write("submodule(my_mod) test\nend submodule\n")
    # Use monkeypatch to spoof the command-line argument.
    monkeypatch.setattr(sys, "argv", ["fparser2", "--std=f2003",
                                      my_file.strpath])
    # Run the relevant script method (main()).
    fparser2.main()
    # Capture the output and check that the code has been output.
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert "File: '" in stderr and "submodule.f90'" in stderr
    assert "Syntax error: at line 1\n>>>submodule(my_mod) test\n" in stderr


def test_main_output_std_f2008(capsys, tmpdir, monkeypatch):
    '''Test that the script main() function parses using the Fortran2008
    standard when --std=f2008. This test attempts to use submodules
    which are supported in Fortran200 so should be parsed correctly.

    '''
    import sys
    # Create a temporary file containing Fortran code to pass into
    # runner().
    my_file = tmpdir.mkdir("sub").join("submodule.f90")
    my_file.write("submodule(my_mod) test\nend submodule\n")
    # Use monkeypatch to spoof the command-line argument.
    monkeypatch.setattr(sys, "argv", ["fparser2", "--std=f2008",
                                      my_file.strpath])
    # Run the relevant script method (main()).
    fparser2.main()
    # Capture the output and check that the code has been output.
    stdout, stderr = capsys.readouterr()
    assert stdout == "SUBMODULE (my_mod) test\nEND SUBMODULE\n"
    assert "File: '" in stderr and "submodule.f90'" in stderr


def test_main_output_std_invalid(capsys, monkeypatch):
    '''Test that the script main() function raises an exception when an
    invalid standard --std=invalid is provided.

    '''
    import sys
    # Use monkeypatch to spoof the command-line argument.
    monkeypatch.setattr(sys, "argv", ["fparser2", "--std=invalid",
                                      "dummy.f90"])
    # Run the relevant script method (main()).
    with pytest.raises(SystemExit) as excinfo:
        fparser2.main()
    assert str(excinfo.value) == "2"
    # Capture the output and check that the code has been output.
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert ("fparser2: error: option --std: invalid choice: 'invalid' "
            "(choose from 'f2003', 'f2008')" in stderr)

# read.py script function runner()

# Create a dummy class (DummyReadArgs) with the required attribute to pass
# into runner() as an argument options class in subsequent tests


class DummyReadArgs(object):
    '''dummy object pretending to be the argument options for the read
    module in scripts.

    :param str task: the value to set the internal task
    variable. Defaults to "show".

    '''
    def __init__(self, task="show"):
        self.task = task


def test_read_runner_nofiles(capsys):
    '''Test that the script deals with no files provided as expected.'''
    # run the relevant script method (runner())
    read.runner(None, DummyArgs(), [])
    # capture the output and check that no output is generated
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert stderr == ""


def test_read_runner_file(capsys, tmpdir):
    '''Test that the script deals with one file as expected.'''
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    read.runner(None, DummyReadArgs(), [str(my_file)])
    stdout, stderr = capsys.readouterr()
    assert "line #1'program hello'" in stdout
    assert "line #2'end program hello'" in stdout
    assert stderr == ""


def test_read_runner_files(capsys, tmpdir):
    '''Test that the script deals with multiple files as expected.'''
    my_file1 = tmpdir.mkdir("sub1").join("hello1.f90")
    my_file1.write("program hello1\nend program hello1\n")
    my_file2 = tmpdir.mkdir("sub2").join("hello2.f90")
    my_file2.write("program hello2\nend program hello2\n")
    read.runner(None, DummyReadArgs(), [str(my_file1), str(my_file2)])
    stdout, stderr = capsys.readouterr()
    assert "line #1'program hello1'" in stdout
    assert "line #2'end program hello1'" in stdout
    assert "line #1'program hello2'" in stdout
    assert "line #2'end program hello2'" in stdout
    assert stderr == ""


def test_read_runner_no_show(tmpdir):
    '''Test that the script raises an exception if the task argument !=
    "show" as that is the only option currently supported.

    '''
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    with pytest.raises(NotImplementedError) as excinfo:
        read.runner(None, DummyReadArgs("invalid"), [str(my_file)])
    assert ("The task option ''invalid'' is invalid. Currently only 'show' "
            "is supported.") in str(excinfo.value)


def test_read_main(capsys, tmpdir, monkeypatch):
    '''Test that the main function works as expected.'''
    import sys
    # Create a temporary file containing Fortran code.
    my_file = tmpdir.mkdir("sub").join("hello.f90")
    my_file.write("program hello\nend program hello\n")
    # Use monkeypatch to spoof the command-line argument
    monkeypatch.setattr(sys, "argv", ["read", str(my_file)])
    # run the relevant script method (main())
    read.main()
    # capture the output and check that the code has been output
    stdout, stderr = capsys.readouterr()
    assert "line #1'program hello'" in stdout
    assert "line #2'end program hello'" in stdout
    assert stderr == ""



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2017-2018 Science and Technology Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
# Modified M.Hambley, UK Met Office
##############################################################################
'''
Test battery associated with fparser.sourceinfo package.
'''
from __future__ import print_function

import os
import tempfile
import pytest
import six

from fparser.common.sourceinfo import FortranFormat, \
                                      get_source_info_str, get_source_info


##############################################################################

def test_format_constructor_faults():
    '''
    Tests that the constructor correctly rejects attempts to create an object
    with "None" arguments.
    '''
    with pytest.raises(Exception):
        _unit_under_test = FortranFormat(True, None)

    with pytest.raises(Exception):
        _unit_under_test = FortranFormat(None, True)


##############################################################################
@pytest.fixture(scope="module",
                params=[(False, False, 'fix', 'Non-strict fixed format'),
                        (False, True, 'f77', 'Strict fixed format'),
                        (True, False, 'free', 'Non-strict free format'),
                        (True, True, 'pyf', 'Strict free format')])
def pretty(request):
    '''
    Returns all possible permutations of flags and their corresponding
    mode and descriptive strings.
    '''
    return request.param


##############################################################################

def test_fortranformat_constructor(pretty):
    #pylint: disable=redefined-outer-name
    '''
    Tests the constructor correctly sets up the object.
    '''
    unit_under_test = FortranFormat(pretty[0], pretty[1])
    assert str(unit_under_test) == pretty[3]
    assert unit_under_test.is_free == pretty[0]
    assert unit_under_test.is_fixed == (not pretty[0])
    assert unit_under_test.is_strict == pretty[1]
    assert unit_under_test.is_f77 == (not pretty[0] and pretty[1])
    assert unit_under_test.is_fix == (not pretty[0] and not pretty[1])
    assert unit_under_test.is_pyf == (pretty[0] and pretty[1])
    assert unit_under_test.mode == pretty[2]


##############################################################################
@pytest.fixture(scope="module",
                params=[(False, False),
                        (False, True),
                        (True, False),
                        (True, True)])
def permutations(request):
    '''
    Returns all possible permutations of the input arguments.
    '''
    return request.param


##############################################################################

def test_fortranformat_equality(permutations, pretty):
    #pylint: disable=redefined-outer-name
    '''
    Tests that the equality operator works as expected.
    '''
    expected = (permutations[0] == pretty[0]) \
        and (permutations[1] == pretty[1])
    unit_under_test = FortranFormat(permutations[0], permutations[1])
    candidate = FortranFormat(pretty[0], pretty[1])
    assert (unit_under_test == candidate) == expected


##############################################################################

def test_fortranformat_invalid():
    '''
    Tests that the equality operator understands that it can't compare apples
    and oranges.
    '''
    unit_under_test = FortranFormat(True, False)
    with pytest.raises(NotImplementedError):
        if unit_under_test == 'oranges':
            raise Exception("That shouldn't have happened")


##############################################################################

@pytest.fixture(scope="module",
                params=[('free', True, False),
                        ('f77', False, True),
                        ('fix', False, False),
                        ('pyf', True, True)])
def mode(request):
    '''
    Returns all possible mode strings and their corresponding flags.
    '''
    return request.param


##############################################################################

def test_fortranformat_from_mode(mode):
    #pylint: disable=redefined-outer-name
    '''
    Tests that the object is correctly created by the from_mode function.
    '''
    unit_under_test = FortranFormat.from_mode(mode[0])
    assert unit_under_test.mode == mode[0]
    assert unit_under_test.is_free == mode[1]
    assert unit_under_test.is_fixed == (not mode[1])
    assert unit_under_test.is_strict == mode[2]
    assert unit_under_test.is_f77 == (not mode[1] and mode[2])
    assert unit_under_test.is_fix == (not mode[1] and not mode[2])
    assert unit_under_test.is_pyf == (mode[1] and mode[2])
    assert str(unit_under_test.mode) == mode[0]


##############################################################################

def test_format_from_mode_bad():
    '''
    Tests that an exception is thrown for an unrecognised mode string.
    '''
    with pytest.raises(NotImplementedError):
        _unit_under_test = FortranFormat.from_mode('cheese')


##############################################################################
# Setting up a pytest fixture in this way is a mechanism for creating
# parameterised tests.
#
# Normally when a test includes a fixture in its argument list the
# corresponding function will be called and the result passed in to the test.
#
# When used like this the test will be called once for each value the fixture
# function returns. So in this case any test including "header" in its
# argument list will be run once with "header" equal to "! -*- fortran -*-",
# then with "header" equal to "! -*- f77 -*-" and so on.
#
# If a test includes multiple parameter fixtures it will be called for every
# permutation thus afforded.
#
@pytest.fixture(scope="module",
                params=[(None, FortranFormat(True, True)),
                        ("! -*- fortran -*-", FortranFormat(False, True)),
                        ("! -*- f77 -*-", FortranFormat(False, True)),
                        ('! -*- f90 -*-', FortranFormat(True, False)),
                        ('! -*- f03 -*-', FortranFormat(True, False)),
                        ('! -*- f08 -*-', FortranFormat(True, False)),
                        ('! -*- fix -*-', FortranFormat(False, False)),
                        ('! -*- pyf -*-', FortranFormat(True, True))])
def header(request):
    '''
    Returns parameters for header tests.
    '''
    return request.param


##############################################################################

_FIXED_SOURCE = '''      program main
      end program main
'''

_FREE_SOURCE = '''program main
end program main
'''

_FIXED_WITH_CONTINUE = '''      program main
          implicit none
          integer :: foo, &
                     bar
      end program main
'''

_FIXED_WITH_COMMENTS = '''!     The program
      program main
c         Enforce explicit variable declaration
          implicit none
*         variables
          integer :: foo
      end program main
'''

# Tabs are not actually in the Fortran character set but fparser has handled
# them in the past even if it shouldn't have. We have to continue handling
# them for the time being until we work out why they were supported.
#
_INITIAL_TAB = "\tprogram main\n"
_MIDDLE_TAB = " \tprogram main\n"


# Another parameterised test fixture. See "header" above.
#
@pytest.fixture(scope="module",
                params=[(None, FortranFormat(False, False)),
                        (_FIXED_SOURCE, FortranFormat(False, False)),
                        (_FREE_SOURCE, FortranFormat(True, False)),
                        (_FIXED_WITH_CONTINUE, FortranFormat(True, False)),
                        (_FIXED_WITH_COMMENTS, FortranFormat(False, False)),
                        (_INITIAL_TAB, FortranFormat(False, False)),
                        (_MIDDLE_TAB, FortranFormat(True, False))])
def content(request):
    '''
    Returns parameters for content tests.
    '''
    return request.param


##############################################################################

def test_get_source_info_str(header, content):
    #pylint: disable=redefined-outer-name
    '''
    Tests that source format is correctly identified when read from a string.
    '''
    full_source = ''
    if header[0] is not None:
        full_source += header[0] + '\n'
    if content[0] is not None:
        full_source += content[0]

    source_info = get_source_info_str(full_source)
    if header[0]:
        assert source_info == header[1]
    else:  # No header
        assert source_info == content[1]


##############################################################################

# Another parameterised test fixture. See "header" above.
#
@pytest.fixture(scope="module",
                params=[('.f', None),
                        ('.f90', None),
                        ('.pyf', FortranFormat(True, True)),
                        ('.guff', None)])
def extension(request):
    '''
    Returns parameters for extension tests.
    '''
    return request.param


##############################################################################

def test_get_source_info_filename(extension, header, content):
    #pylint: disable=redefined-outer-name
    '''
    Tests that source format is correctly identified when read from a file.
    '''
    full_source = ''
    if header[0] is not None:
        full_source += header[0] + '\n'
    if content[0] is not None:
        full_source += content[0]

    source_file, filename = tempfile.mkstemp(suffix=extension[0], text=True)
    os.close(source_file)

    with open(filename, 'w') as source_file:
        print(full_source, file=source_file)

    try:
        source_info = get_source_info(filename)
        if extension[1] is not None:
            assert source_info == extension[1]
        elif header[0] is not None:
            assert source_info == header[1]
        else:  # No header
            assert source_info == content[1]
    except Exception as ex:
        os.remove(filename)
        raise ex


##############################################################################

def test_get_source_info_file(extension, header, content):
    #pylint: disable=redefined-outer-name
    '''
    Tests that source format is correctly identified when read from a file.
    '''
    full_source = ''
    if header[0] is not None:
        full_source += header[0] + '\n'
    if content[0] is not None:
        full_source += content[0]

    with tempfile.TemporaryFile(mode='w+',
                                suffix=extension[0]) as source_file:
        print(full_source, file=source_file)
        source_file.seek(0)

        source_info = get_source_info(source_file)
        if header[0] is not None:
            assert source_info == header[1]
        else:  # No header
            assert source_info == content[1]


def test_get_source_info_utf8():
    '''
    Tests that Fortran code containing a unicode character can be read
    by the get_source_info method.

    '''
    encoding = dict(encoding='UTF-8') if six.PY3 else {}
    with tempfile.NamedTemporaryFile(mode='w', **encoding) as tmp_file:
        content = u'''
            ! A fortran comment with a unicode character "{0}"
        '''.format(u"\u2014")
        if six.PY2:
            content = content.encode('UTF-8')
        tmp_file.write(content)
        tmp_file.flush()

        source_info = get_source_info(tmp_file.name)
    assert source_info is not None


##############################################################################

def test_get_source_info_wrong():
    '''
    Tests that get_source_info throws an exception if passed the wrong type
    of argument.
    '''
    with pytest.raises(ValueError):
        _source_info = get_source_info(42)  # Obviously wrong

    with pytest.raises(ValueError):
        _source_info = get_source_info(['one'])  # Less obviously wrong

##############################################################################
